# flake8: noqa
from __future__ import absolute_import
from webfriend.rpc.reply import Reply
from webfriend.rpc.base import Base
from webfriend.rpc.browser import Browser
from webfriend.rpc.console import Console
from webfriend.rpc.dom import DOM, DOMElement
from webfriend.rpc.emulation import Emulation
from webfriend.rpc.event import Event
from webfriend.rpc.input import Input
from webfriend.rpc.inspector import Inspector
from webfriend.rpc.network import Network
from webfriend.rpc.overlay import Overlay
from webfriend.rpc.page import Page
from webfriend.rpc.runtime import Runtime
from webfriend.rpc.schema import Schema
from webfriend.rpc.target import Target



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.commands.base import CommandProxy
from webfriend import exceptions


class CookiesProxy(CommandProxy):
    def all(self, urls=None):
        """
        Return a list of all cookies, optionally restricted to just a specific URL.

        #### Arguments

        - **urls** (`list`, optional):
            If specified, this is a list of URLs to retrieve cookies for.

        #### Returns
        A `list` of `dict` objects containing definitions for each cookie.
        """
        return [
            c.as_dict() for c in self.tab.network.get_cookies(urls)
        ]

    def query(self, name=None, **filters):
        """
        Query all known cookies and return a list of cookies matching those with specific values.

        #### Arguments

        The first argument (optional) is the name of the cookie as defined in its description. All
        options fields are taken as additional filters used to further restrict which cookies are
        returned.

        - **value** (`str`, optional):

            The value of the cookie.

        - **domain** (`str`, optional):

            The domain for which the cookie is valid for.

        - **path** (`str`, optional):

            The path valid of the cookie.

        - **expires** (`int`, optional):

            Cookie expiration date as the number of seconds since the UNIX epoch.

        - **size** (`int`, optional):

            The size of the cookie, in bytes.

        - **httpOnly** (`bool`, optional):

            Whether the cookie is marked as "HTTP only" or not.

        - **secure** (`bool`, optional):

            Whether the cookie is marked as secure or not.

        - **session** (`bool`, optional):

            Whether the cookie is marked as a session cookie or not.

        - **sameSite** (`bool`, optional):

            Whether the cookie is marked as a sameSite cookie or not.

        #### Returns
        A `list` of `dicts` describing the cookies that matched the given query, whose fields
        will be the same as the ones described above.
        """
        if name is not None:
            filters['name'] = name

        def _filter(cookie):
            for k, v in filters.items():
                if v is not None:
                    if cookie.get(k) != v:
                        return False

            return True

        return [
            c for c in self.all() if _filter(c)
        ]

    def get(self, name, domain=None):
        """
        Retrieve a specific cookie by name and (optionally) domain.  The domain should be provided
        to ensure the cookie returns at most one result.

        #### Arguments

        - **domain** (`str`, optional):

            The domain value of the cookie being retrieved to ensure an unambiguous result.

        #### Returns
        A `dict` describing the cookie returned, or `None`.

        #### Raises
        A `webfriend.exceptions.TooManyResults` exception if more than one cookie matches the given
        values.
        """
        results = self.query(name=name, domain=domain)

        if not len(results):
            return None
        elif len(results) == 1:
            return results[0]
        else:
            raise exceptions.TooManyResults(
                "Cookie name '{}' is ambiguous, {} cookies matched. Provide a more specific filter"
                " using {}::query".format(
                    name,
                    len(results),
                    self.as_qualifier()
                )
            )

    def delete(self, name, domain=None):
        """
        Delete the cookie specified by the given **name** and (optionally) **domain**.

        #### Arguments

        - **name** (`str`, optional):

            The name of the cookie to delete.

        - **domain** (`str`, optional):

            The domain value of the cookie being retrieved to ensure an unambiguous result.
        """
        if domain is None:
            cookie_by_name = self.get(name)
            domain = cookie_by_name.get('domain')

        self.tab.network.delete_cookie(domain, name)

    # def set_cookie(
    #     self,
    #     url,
    #     name,
    #     value,
    #     domain=None,
    #     path=None,
    #     secure=None,
    #     http_only=None,
    #     same_site=None,
    #     expires=None
    # ):
    def set(self, name, **kwargs):
        """
        Create or update a cookie based on the given values.

        #### Arguments

        - **name** (`str`):

            The name of the cookie to set.

        - **value** (any):

            The value to set in the cookie.

        - **url** (`str`, optional):

            The URL to associate the cookie with. This is important when dealing with things like
            host-only cookies (if **domain** isn't set, a host-only cookie will be created.)  In
            this case, the cookie will only be valid for the exact URL that was used.

            The default value is the URL of the currently active tab.

        - **domain** (`str`, optional):

            The domain for which the cookie will be presented.

        - **path** (`str`, optional):

            The path value of the cookie.

        - **secure** (`bool`, optional):

            Whether the cookie is flagged as secure or not.

        - **http_only** (`bool`, optional):

            Whether the cookie is flagged as an HTTP-only cookie or not.

        - **same_site** (`str`, optional):

            Sets the "Same Site" attribute of the cookie.  The value "strict" will restrict any
            cross-site usage of the cookie.  The value "lax" allows top-level navigation changes
            to receive the cookie.

        - **expires** (`int`, optional):

            Specifies when the cookie expires in epoch seconds (number of seconds
            since 1970-01-01 00:00:00 UTC).
        """
        if not kwargs.get('url', None):
            kwargs['url'] = self.tab.url

        if kwargs.get('same_site', None):
            kwargs['same_site'] = kwargs['same_site'].title()

        if not kwargs.get('value', None):
            raise ValueError("'value' option must be specified")

        return self.tab.network.set_cookie(**kwargs)



from __future__ import absolute_import
from webfriend.rpc import Base


class Overlay(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Overlay
    """
    domain = 'Overlay'
    supports_events = False

    default_highlight_color = {
        'r': 0,
        'g': 99,
        'b': 193,
        'a': 0.95,
    }

    default_highlight_outline_color = {
        'r': 255,
        'g': 255,
        'b': 255,
        'a': 0,
    }

    show_rulers = False
    show_extension_lines = False
    show_info_tooltip = False

    def show_paint_rectangles(self):
        self.call('setShowPaintRects', result=True)

    def hide_paint_rectangles(self):
        self.call('setShowPaintRects', result=False)

    def show_debug_borders(self):
        self.call('setShowDebugBorders', show=True)

    def hide_debug_borders(self):
        self.call('setShowDebugBorders', show=False)

    def show_fps_counter(self):
        self.call('setShowFPSCounter', show=True)

    def hide_fps_counter(self):
        self.call('setShowFPSCounter', show=False)

    def show_scroll_bottlenecks(self):
        self.call('setShowScrollBottleneckRects', show=True)

    def hide_scroll_bottlenecks(self):
        self.call('setShowScrollBottleneckRects', show=False)

    def show_size_on_resize(self):
        self.call('setShowViewportSizeOnResize', show=True)

    def hide_size_on_resize(self):
        self.call('setShowViewportSizeOnResize', show=False)

    def suspend(self):
        self.call('setSuspended', suspended=True)

    def resume(self):
        self.call('setSuspended', suspended=False)

    def enable_inspect_mode(self):
        self.call('setInspectMode', mode='searchForNode')

    def disable_inspect_mode(self):
        self.call('setInspectMode', mode='none')

    def highlight_rect(self, x, y, width, height, color=None, outline_color=None):
        if not isinstance(color, dict):
            color = self.default_highlight_color

        if not isinstance(outline_color, dict):
            outline_color = self.default_highlight_outline_color

        self.call(
            'highlightRect',
            x=x,
            y=y,
            width=width,
            height=height,
            color=color,
            outlineColor=outline_color
        )

    def highlight_node(
        self,
        node_id=None,
        backend_node_id=None,
        object_id=None,
        show_rulers=None,
        show_extension_lines=None,
        show_info_tooltip=None,
        content_color=None,
        padding_color=None,
        border_color=None,
        margin_color=None,
        event_target_color=None,
        shape_color=None,
        shape_margin_color=None,
    ):
        params = {
            'highlightConfig': {},
        }

        if node_id:
            params['nodeId'] = node_id
        elif backend_node_id:
            params['backendNodeId'] = backend_node_id
        elif object_id:
            params['objectId'] = object_id
        else:
            raise ValueError("Must specify one of node_id or object_id")

        if not show_rulers:
            params['highlightConfig']['showRulers'] = self.show_rulers

        if not show_info_tooltip:
            params['highlightConfig']['showInfo'] = self.show_info_tooltip

        if not show_extension_lines:
            params['highlightConfig']['showExtensionLines'] = self.show_extension_lines

        if isinstance(content_color, dict):
            params['highlightConfig']['contentColor'] = content_color
        else:
            params['highlightConfig']['contentColor'] = self.default_highlight_color

        if isinstance(padding_color, dict):
            params['highlightConfig']['paddingColor'] = padding_color

        if isinstance(border_color, dict):
            params['highlightConfig']['borderColor'] = border_color

        if isinstance(margin_color, dict):
            params['highlightConfig']['marginColor'] = margin_color

        if isinstance(event_target_color, dict):
            params['highlightConfig']['eventTargetColor'] = event_target_color

        if isinstance(shape_color, dict):
            params['highlightConfig']['shapeColor'] = shape_color

        if isinstance(shape_margin_color, dict):
            params['highlightConfig']['shapeMarginColor'] = shape_margin_color

        self.call('highlightNode', **params)

    def hide_highlight(self):
        self.call('hideHighlight')



class WebfriendError(Exception):
    pass


class ProtocolError(WebfriendError):
    pass


class TimeoutError(WebfriendError):
    pass


class NotFound(WebfriendError):
    pass


class NetworkError(WebfriendError):
    pass


class HttpError(WebfriendError):
    pass


class EmptyResult(WebfriendError):
    pass


class TooManyResults(WebfriendError):
    pass


class NotImplemented(WebfriendError):
    pass



from __future__ import absolute_import
from webfriend.rpc import Base
from webfriend import exceptions, utils
from base64 import b64decode
import os
import time
import logging


class Page(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Page
    """
    domain = 'Page'
    capture_formats = ['jpeg', 'png']

    def navigate(self, url, referrer=None, transition_type=None, clear_request_cache=True):
        params = {
            'url': url,
        }

        if referrer:
            params['referrer'] = referrer

        if transition_type:
            params['transitionType'] = transition_type

        # clear the network request cache between explicit page loads
        if clear_request_cache is True:
            self.tab.reset_network_request_cache()

        init_frame_id = self.call('navigate', **params).get('frameId')

        # block until the network first request finished (success or failure)
        self.tab.wait_for('Network.loadingFinished')

        # get the request corresponding to the frameID (this is the request for the given URL)
        net_request = self.tab.get_network_request(init_frame_id)

        if net_request:
            if net_request.get('success') is True:
                response = net_request.get('response', {})
                status = response['status']

                if status < 400:
                    return response
                else:
                    exc = exceptions.HttpError('HTTP {}'.format(status))
                    exc.response = response
                    raise exc

            else:
                pre_event = net_request.get('before', {})
                fail_event = net_request.get('response', {})
                msg = fail_event.get('errorText', 'UNKNOWN')
                exc = exceptions.NetworkError(msg)
                exc.event = fail_event
                exc.url = pre_event.get('documentURL')
                raise exc

    def reload(self, ignore_cache=False, eval_on_load=None):
        params = {
            'ignoreCache': ignore_cache,
        }

        if isinstance(eval_on_load, basestring):
            params['scriptToEvaluateOnLoad'] = eval_on_load

        self.call('reload', **params)

    def stop(self):
        self.call('stopLoading')

    def dialog(self, accept, text=None):
        params = {
            'accept': accept,
        }

        if isinstance(text, basestring):
            params['promptText'] = text

        self.call('handleJavaScriptDialog', **params)

    def capture_screenshot(self, destination, format=None, quality=None, from_surface=True, reply_timeout=None):
        params = {}

        if format is not None:
            if format in self.capture_formats:
                params['format'] = format
            else:
                raise AttributeError("Invalid format, must be one of: {}".format(
                    ', '.join(self.capture_formats)
                ))

        if format == 'jpeg' and isinstance(quality, int):
            params['quality'] = quality

        if from_surface is False:
            params['fromSurface'] = False

        reply = self.call('captureScreenshot', reply_timeout=reply_timeout, **params)

        if isinstance(destination, basestring):
            data = open(destination, 'w')

        body = b64decode(reply.get('data'))
        data.write(body)

        return len(body)

    def start_screencast(
        self,
        format='png',
        jpeg_quality=85,
        max_width=None,
        max_height=None,
        every_nth=None,
        duration=None,
        destination=None,
        filename_format=None,
        additional_values=None
    ):
        if hasattr(self, '_active_screencast'):
            if self._active_screencast:
                raise Exception("There is already an active screencast session running.")
        else:
            self._active_screencast = None

        if not filename_format:
            filename_format = '{page_id}_{screencast_id}_{frame_number}.{extension}'

        params = {
            'format': format,
        }

        if not additional_values:
            additional_values = {}

        if format == 'png':
            additional_values['extension'] = 'png'
        elif format == 'jpeg':
            additional_values['extension'] = 'jpg'

            if jpeg_quality:
                params['quality'] = jpeg_quality

        if max_width:
            params['maxWidth'] = max_width

        if max_height:
            params['maxHeight'] = max_height

        if every_nth:
            params['everyNthFrame'] = every_nth

        if destination:
            abs_dir = os.path.abspath(destination)

            if not os.path.isdir(abs_dir):
                os.makedirs(abs_dir)

            # generate the frame handler function for this screencast
            handler = self.get_receive_screencast_frame_handler(
                utils.random_string(8),
                abs_dir,
                filename_format,
                **{
                    'duration':                 duration,
                    'poll_interval_multiplier': (every_nth or 1),
                    'extra':                    additional_values,
                }
            )

            # register the event handler
            self._active_screencast = self.on('screencastFrame', handler)

        # start the screencast
        self.call('startScreencast', **params)

        # wait for the screencast to start before returning
        self.tab.wait_for('Page.screencastVisibilityChanged')

        # note the time
        self._active_screencast_started_at = time.time()

        return {
            'started_at': self._active_screencast_started_at,
            'handler_id': self._active_screencast,
        }

    def get_receive_screencast_frame_handler(
        self,
        screencast_id,
        destination,
        filename_format,
        started_at=None,
        duration=None,
        poll_interval_multiplier=1,
        extra=None
    ):
        self._active_screencast_frame_count = 0

        def handle(frame):
            frame_session_id = frame.get('sessionId')

            filename = filename_format.format(**{
                'page_id':       self.tab.frame_id.replace('-', ''),
                'screencast_id': str(screencast_id),
                'frame_number':  self._active_screencast_frame_count,
                'metadata':      frame.get('metadata'),
                'extension':     extra.get('extension'),
                'extra':         (extra if isinstance(extra, dict) else {}),
            })

            # write the data out to the file
            with open(os.path.join(destination, filename), 'w') as file:
                data = frame.get('data')

                if data and len(data):
                    file.write(b64decode(data))

            # acknowledge the frame
            self.screencast_frame_ack(frame_session_id)
            logging.debug('Wrote frame {} to {}'.format(self._active_screencast_frame_count, filename))

            self._active_screencast_frame_count += 1

            # if we only want to screencast for a specific duration, check the
            # times and stop if necessary
            if duration:
                if hasattr(self, '_active_screencast_started_at'):
                    if time.time() > (self._active_screencast_started_at + (duration / 1e3)):
                        self.stop_screencast()

        return handle

    def stop_screencast(self, timeout=10000):
        if self._active_screencast:
            logging.debug('Stopping screencast {}'.format(self._active_screencast))
            self.remove_handler(self._active_screencast)
            self._active_screencast = None

        self.call('stopScreencast')
        return True

    def screencast_frame_ack(self, session_id):
        self.call('screencastFrameAck', sessionId=session_id)
        return True

    # addScriptToEvaluateOnLoad[ scriptSource=string ]
    #   -> string(identifier)
    #
    # removeScriptToEvaluateOnLoad
    #   -> string(identifier)
    #
    # setAutoAttachToCreatedPages [ autoAttach=bool ] -> Nothing
    #
    # getNavigationHistory
    #   -> integer(currentIndex)
    #   -> []entries
    #      @NavigationEntry
    #       id=integer           Unique id of the navigation history entry.
    #       url=string           URL of the navigation history entry.
    #       userTypedURL=string  URL that the user typed in the url bar.
    #       title=string         Title of the navigation history entry.
    #       transitionType=(
    #           link,
    #           typed,
    #           auto_bookmark,
    #           auto_subframe,
    #           manual_subframe,
    #           generated,
    #           auto_toplevel,
    #           form_submit,
    #           reload,
    #           keyword,
    #           keyword_generated,
    #           other
    #       )
    #
    # navigateToHistoryEntry [ entryId=integer ]
    #
    # getResourceTree
    # getResourceContent
    # searchInResource
    #
    # setDocumentContent [ frameId=string html=string ]



from __future__ import absolute_import
from webfriend.rpc import Base


class Browser(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Browser
    """

    domain = 'Browser'
    supports_events = False

    def get_window_size(self):
        # response = self.call('getWindowBounds')
        pass



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.commands.base import CommandProxy
from webfriend.utils import autotype
from webfriend.scripting.parser import types


class FormatProxy(CommandProxy):
    qualifier = 'fmt'
    doc_name = 'Format'

    def autotype(self, value):
        """
        Takes a string value and attempts to automatically determines the data type.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        #### Returns
        The value converted to the automatically-detected type, or the original unmodified value if
        a type could not be determined.
        """
        return autotype(value)

    def strip(self, value, characters=None):
        """
        Removes the whitespace or a set of given characters from the leading and trailing ends of
        the given string.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        - **characters** (`str`, optional):

            If specified, any of these characters will be trimmed from the string instead of
            whitespace.

        #### Returns
        The trimmed string.
        """
        if isinstance(value, basestring):
            return value.strip(characters)
        return value

    def lstrip(self, value, characters=None):
        """
        Removes the whitespace or a set of given characters from the leading end of the
        given string.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        - **characters** (`str`, optional):

            If specified, any of these characters will be trimmed from the string instead of
            whitespace.

        #### Returns
        The trimmed string.
        """
        if isinstance(value, basestring):
            return value.lstrip(characters)
        return value

    def rstrip(self, value, characters=None):
        """
        Removes the whitespace or a set of given characters from the trailing end of the
        given string.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        - **characters** (`str`, optional):

            If specified, any of these characters will be trimmed from the string instead of
            whitespace.

        #### Returns
        The trimmed string.
        """
        if isinstance(value, basestring):
            return value.rstrip(characters)
        return value

    def lower(self, value):
        """
        Returns the given string with all characters replaced with their lowercase variants.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        #### Returns
        The lowercased string.
        """
        if isinstance(value, basestring):
            return value.lower()
        return value

    def upper(self, value):
        """
        Returns the given string with all characters replaced with their uppercase variants.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        #### Returns
        The uppercased string.
        """
        if isinstance(value, basestring):
            return value.upper()
        return value

    def title(self, value):
        """
        Returns the given string with the first character of each whitespace-separated word
        capitalized.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        #### Returns
        The title-cased string.
        """
        if isinstance(value, basestring):
            return value.title()
        return value

    def replace(self, value, find, replace, count=None):
        """
        Locates a given string or pattern in the given string and replaces occurrences with the
        given replacement value.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        - **find** (`str`, `regex`):

            The exact string or regular expression that will match the given **value**.

        - **replace** (`str`):

            The string (or string pattern for regular expression) that will replace occurrences in
            the given **value**.

            If **find** is a regular expression, then **replace** is interpreted as follows:

            - Groups in the regular expression can be referenced numerically with backreferences
              like `\\1`, `\\3`.  The number following the backslash refers to the _nth_ group in
              the **find** expression, where group numbering starts from 1.  The `\\0`
              backreference refers to the entire matched substring.

            - Named captures in **find** can be referenced such that `\\g<name>` refers to named
              capture group `name` in **find** (which would be specified as `(?P<name>...)`.)

        - **count** (`int`, optional):

            If specified, up to this many occurrences will be matched before matching stops.  Use
            this to limit the number of replacement operations that can occur.

        #### Returns
        The resulting string with all replacements (if any) applied.
        """
        if isinstance(find, types.RegularExpression):
            if isinstance(count, int):
                return find.rx.sub(replace, value, count=count)
            else:
                return find.rx.sub(replace, value)
        else:
            return value.replace(find, replace, count)

    def split(self, value, on, count=None):
        """
        Split an input string into a list of strings.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        - **on** (`str`):

            The character or substring to split **value** on.  This value will be
            discarded from the resulting list of strings.

        - **count** (`int`, optional):

            If specified, only perform up to this many splits before returning the remaining
            unsplit string as the final element in the list.

        #### Returns
        A `list` of at least 1 element.
        """
        return value.split(on, count)

    def rsplit(self, value, on, count=None):
        """
        Split an input string into a list of strings, starting from the right-hand side.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        - **on** (`str`):

            The character or substring to split **value** on.  This value will be
            discarded from the resulting list of strings.

        - **count** (`int`, optional):

            If specified, only perform up to this many splits before returning the remaining
            unsplit string as the final element in the list.

        #### Returns
        A `list` of at least 1 element.
        """
        return value.rsplit(on, count)

    def join(self, values, joiner=','):
        """
        Join a list of values into a single string.

        #### Arguments

        - **values** (`list`):

            A list of values (of any type) to stringify and join by a given joiner string.

        - **joiner** (`str`):

            The string to join list elements by.

        #### Returns
        The resulting joined string.
        """
        if not isinstance(values, (list, tuple)):
            values = [values]

        values = ['{}'.format(v) for v in values]

        return joiner.join(values).encode('UTF-8')



from __future__ import absolute_import
from webfriend.rpc import Base
import math


class Emulation(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Emulation
    """
    domain = 'Emulation'
    supports_events = False

    valid_orientations = [
        'portraitPrimary',
        'portraitSecondary',
        'landscapePrimary',
        'landscapeSecondary',
    ]

    def set_device_metrics_override(
        self,
        width=0,
        height=0,
        device_scale_factor=0,
        mobile=False,
        fit_window=False,
        mobile_screen_width=0,
        mobile_screen_height=0,
        mobile_position_x=0,
        mobile_position_y=0,
        screen_orientation_type=None,
        screen_orientation_angle=None
    ):
        params = {
            'width':             int(math.ceil(width)),
            'height':            int(math.ceil(height)),
            'deviceScaleFactor': device_scale_factor,
            'mobile':            mobile,
            'fitWindow':         fit_window,
        }

        if mobile is True:
            params['screenWidth']  = mobile_screen_width
            params['screenHeight'] = mobile_screen_height
            params['positionX']    = mobile_position_x
            params['positionY']    = mobile_position_y

        if screen_orientation_type is not None:
            params['screenOrientation'] = {
                'type':  screen_orientation_type,
                'angle': (screen_orientation_angle or 0),
            }

        self.call('setDeviceMetricsOverride', **params)

    def clear_device_metrics_override(self):
        self.call('clearDeviceMetricsOverride')

    def force_viewport(self, x=0, y=0, scale=1.0):
        self.call('forceViewport', x=x, y=y, scale=scale)

    def reset_viewport(self):
        self.call('resetViewport')

    def set_page_scale_factor(self, factor):
        self.call('setPageScaleFactor', pageScaleFactor=factor)

    def reset_page_scale_factor(self):
        self.call('resetPageScaleFactor')

    def set_visible_size(self, width, height):
        self.call(
            'setVisibleSize',
            width=int(math.ceil(width)),
            height=int(math.ceil(height))
        )

    def enable_script_execution(self):
        self.call('setScriptExecutionDisabled', value=False)

    def disable_script_execution(self):
        self.call('setScriptExecutionDisabled', value=True)

    def set_geolocation_override(self, latitude, longitude, accuracy):
        self.call(
            'setGeolocationOverride',
            latitude=latitude,
            longitude=longitude,
            accuracy=accuracy
        )

    def clear_geolocation_override(self):
        self.call('clearGeolocationOverride')

    def enable_touch_emulation(self, configuration=None):
        params = {
            'enabled': True,
        }

        if configuration is not None:
            params['configuration'] = configuration

        self.call('setTouchEmulationEnabled', **params)

    def set_emulated_media(self, media):
        self.call('setEmulatedMedia', media=media)

    def set_cpu_throttling_rate(self, rate):
        self.call('setCPUThrottlingRate', rate=rate)

    def can_emulate(self):
        return self.call_boolean_response('canEmulate')

    def set_default_background_color_override(self, red, green, blue, alpha=1.0):
        self.call('setDefaultBackgroundColorOverride', color={
            'r': red,
            'g': green,
            'b': blue,
            'a': alpha,
        })

    def clear_default_background_color_override(self):
        self.call('setDefaultBackgroundColorOverride')



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.parser import MetaModel, to_value, variables


class LoopBlock(MetaModel):
    def execute_loop(self, environment, scope=None):
        if scope is None:
            scope = environment.scope

        i = 0

        if self.iterable:
            loop = self.iterable

            if isinstance(loop.iterator, variables.Variable):
                iter_value = to_value(loop.iterator, scope)
            else:
                _, iter_value = environment.execute(loop.iterator, scope=scope)

            try:
                if isinstance(iter_value, dict):
                    iter_value = iter_value.items()

                if not iter_value:
                    return

                for item in iter(iter_value):
                    if len(loop.variables) > 1 and isinstance(item, tuple):
                        # unpack iter items into the variables we were given
                        for var_i, var in enumerate(loop.variables[0:len(item)]):
                            if not var.skip:
                                if var_i < len(item):
                                    scope.set(var.as_key(scope), item[var_i], force=True)

                        # null out any remaining variables
                        if len(loop.variables) > len(item):
                            for var in loop.variables[len(item):]:
                                scope.set(var.as_key(scope), None, force=True)

                    else:
                        scope.set(loop.variables[0].as_key(scope), item, force=True)

                    scope.set('index', i, force=True)

                    for result in self.iterate_blocks(i, environment, scope):
                        yield result

                    i += 1

            except TypeError:
                raise TypeError("Cannot loop on result of {}".format(loop.iterator))

        elif self.bounded:
            loop = self.bounded
            environment.execute(loop.initial, scope=scope)

            while loop.termination.evaluate(environment, scope=scope):
                scope.set('index', i, force=True)

                for result in self.iterate_blocks(i, environment, scope):
                    yield result

                environment.execute(loop.next, scope=scope)
                i += 1

        elif self.truthy:
            loop = self.truthy

            while loop.termination.evaluate(environment, scope=scope):
                scope.set('index', i, force=True)

                for result in self.iterate_blocks(i, environment, scope):
                    yield result

                i += 1

        elif self.fixedlen:
            loop = self.fixedlen
            count = int(to_value(loop.count, scope))

            for _ in range(count):
                scope.set('index', i, force=True)

                for result in self.iterate_blocks(i, environment, scope):
                    yield result

                i += 1

        else:
            while True:
                scope.set('index', i, force=True)

                for result in self.iterate_blocks(i, environment, scope):
                    yield result

                i += 1

    def iterate_blocks(self, i, environment, scope):
        for subblock in self.blocks:
            yield i, environment, subblock, scope


class FlowControlWord(MetaModel):
    pass


class FlowControlMultiLevel(Exception):
    def __init__(self, message, levels=1, **kwargs):
        self.levels = levels
        super(FlowControlMultiLevel, self).__init__(message, **kwargs)


class FlowControlBreak(FlowControlMultiLevel):
    pass


class FlowControlContinue(FlowControlMultiLevel):
    pass



from __future__ import absolute_import
import copy
import json
import logging
import os
import random
import requests
import shutil
import tempfile
import time
from webfriend.tab import Tab
from webfriend.scripting.execute import execute_script
from webfriend.utils.commands import locate_browser_process
from ephemeral_port_reserve import reserve, LOCALHOST
import subprocess
from urlparse import urlparse
from collections import OrderedDict

DEFAULT_DEBUGGER_URL = 'http://localhost:9222'


class Chrome(object):
    browser_arguments = [
        '--headless',
        '--disable-gpu',
        '--hide-scrollbars',
    ]

    temp_profile_preferences = {
        'homepage': 'about:blank',
        'homepage_is_newtabpage': True,
        'browser': {
            'show_home_button': False,
        },
        'session': {
            'restore_on_startup': 1,
            'startup_urls': [
                'about:blank',
            ],
        },
        'bookmark_bar': {
            'show_on_all_tabs': False,
        },
        'sync_promo': {
            'show_on_first_run_allowed': False,
        },
        'distribution': {
            'do_not_create_any_shortcuts': True,
            'do_not_create_desktop_shortcut': True,
            'do_not_create_quick_launch_shortcut': True,
            'do_not_create_taskbar_shortcut': True,
            'do_not_launch_chrome': True,
            'do_not_register_for_update_launch': True,
            'import_bookmarks': False,
            'import_history': False,
            'import_home_page': True,
            'import_search_engine': False,
            'make_chrome_default': False,
            'make_chrome_default_for_user': False,
            'ping_delay': 60,
            'require_eula': False,
            'suppress_first_run_bubble': True,
            'suppress_first_run_default_browser_prompt': True,
            'system_level': False,
            'verbose_logging': False,
        },
    }

    def __init__(
        self,
        debug_url=None,
        ping_retries=160,
        ping_delay=125,
        use_temp_profile=True,
        foreground=False
    ):
        self.temp_profile_path = None
        self.args = copy.copy(self.browser_arguments)

        if foreground is True or os.getenv('WEBFRIEND_DEBUG', '').lower() in ['1', 'true']:
            self.debug_url = (debug_url or DEFAULT_DEBUGGER_URL)
            self.args.remove('--headless')
        else:
            self.debug_url = debug_url

        self._process = None
        self.tabs = OrderedDict()
        self.background = None
        self.default_tab = None
        self.ping_retries = ping_retries
        self.ping_delay = ping_delay
        self.use_temp_profile = use_temp_profile
        self.started_at = time.time()

    def start(self):
        retries = 0
        self.started_at = time.time()

        if self.debug_url is None:
            port = reserve()
            self.debug_url = 'http://{}:{}'.format(LOCALHOST, port)

        # if we can already ping the given debug url, then stop here and connect to that
        if self.ping():
            self.sync()
            self.postlaunch()
            return self

        else:
            if self.use_temp_profile:
                # otherwise, we're going to launch the process ourselves
                self.create_temp_profile()

            self.args = list(
                self.args + [
                    '--remote-debugging-port={}'.format(
                        urlparse(self.debug_url).port
                    ),
                ]
            )

            self.launch_browser(self.args)

            while True:
                if self.ping():
                    logging.info('Browser process running in {}ms, pid={}'.format(
                        int((time.time() - self.started_at) * 1e3),
                        self._process.pid
                    ))

                    self.sync()
                    self.postlaunch()

                    return self

                if retries >= self.ping_retries:
                    raise Exception("Failed to connect to browser RPC after {} attempts".format(retries))

                retries += 1
                time.sleep(self.ping_delay / 1e3)

    def stop(self):
        try:
            for tab_id, tab in self.tabs.items():
                logging.info('Stopping tab {}'.format(tab_id))
                tab.stop()

            if self._process:
                logging.info('Stopping browser process pid={}'.format(self._process.pid))
                self._process.terminate()
                self._process.wait()
                logging.info('Browser exited with status {}'.format(self._process.returncode))
                logging.info('Total run time: {}ms'.format(
                    int((time.time() - self.started_at) * 1e3)
                ))

        finally:
            self.destroy_temp_profile()
            self._devnull.close()
            self._devnull = None
            self._process = None

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def ping(self):
        try:
            requests.head('{}/json'.format(self.debug_url))
            return True
        except requests.exceptions.ConnectionError:
            return False

    def sync(self, wait_for_tab=10000, interval=250):
        tabs_seen = set()
        default_tab = None
        started_at = time.time()
        tabs = []

        while time.time() < (started_at + (wait_for_tab / 1e3)):
            manifest = requests.get('{}/json'.format(
                self.debug_url)
            ).json()

            tabs = [
                t for t in manifest
            ]

            if len(tabs):
                break

            time.sleep(interval / 1e3)

        if not len(tabs):
            raise Exception("Browser creation failed")

        for tab in tabs:
            frame_id = tab.get('id')

            if frame_id is not None:
                tabs_seen.add(frame_id)

                if tab['type'] == 'page':
                    if frame_id in self.tabs:
                        self.tabs[frame_id].description.update(tab)

                    else:
                        logging.info('Register tab {}'.format(frame_id))
                        self.tabs[frame_id] = Tab(self, tab, frame_id=frame_id)

                        if default_tab is None:
                            default_tab = frame_id

        for k in self.tabs.keys():
            if k not in tabs_seen:
                del self.tabs[k]

        if default_tab is not None:
            self.default_tab = default_tab

        if not self.background:
            self.background = self.default

    def launch_browser(self, arguments=[]):
        process_path = locate_browser_process()
        logging.info('Launching browser process: {} {}'.format(
            process_path,
            ' '.join(arguments)
        ))

        self._devnull = open(os.devnull, 'w')
        self._process = subprocess.Popen(
            [process_path] + arguments,
            stdout=self._devnull,
            stderr=self._devnull
        )

    def postlaunch(self):
        if self.background:
            self.background.target.enable_auto_attach()
            self.background.target.enable_attach_to_frames()
            # self.background.target.enable_discover_targets()

    @property
    def default(self):
        if self.default_tab is None:
            raise Exception("No default tab available")

        if self.default_tab in self.tabs:
            return self.tabs[self.default_tab]
        else:
            raise Exception("Cannot find tab '{}'".format(self.default_tab))

    def create_tab(self, url, width=None, height=None):
        reply = self.background.target.create_target(url, width=width, height=height)
        target_id = reply.get('targetId')

        self.sync()
        return target_id

    def switch_to_tab(self, id):
        """
        Switches the active tab to a given tab.

        #### Arguments

        - **id** (`str`):
            The ID of the tab to switch to, or one of a selection of special values:

            #### Special Values

            - `next`: Switch to the next tab in the tab sequence (in the UI, this is the tab to
               the immediate _right_ of the current tab.)  If the current tab is the last in the
               sequence, wrap-around to the first tab.

            - `previous`: Switch to the previous tab in the tab sequence (in the UI, this is the
               tab to the immediate _left_ of the current tab.)  If the current tab is the first in
               the sequence, wrap-around to the last tab.

            - `first`: Switch to the first tab.

            - `last`: Switch to the last tab.

            - `random`: Switch to a randomly selected tab.

            - **Negative integer**: Switch to the tab _n_ tabs to the left of the current tab (e.g.
              if you specify `-2`, then switch to the tab two to the left of the current one.)

            - **Positive integer**: Switch to the _nth_ tab.  Tabs are counted from left-to-right,
              starting from zero (`0`).

        #### Returns
        An instance of `webfriend.Tab` representing the tab that was just switched to.
        """
        tab_ids = self.tabs.keys()
        current = tab_ids.index(self.default_tab)

        if current >= 0 and len(tab_ids):
            if id == 'next':
                id = tab_ids[(current + 1) % len(tab_ids)]
            elif id == 'previous':
                id = tab_ids[(current - 1) % len(tab_ids)]
            elif id == 'first':
                id = tab_ids[0]
            elif id == 'last':
                id = tab_ids[len(tab_ids) - 1]
            elif id == 'random':
                id = tab_ids[random.random(0, len(tab_ids))]
            elif isinstance(id, (int, float)):
                if id < 0:
                    id = tab_ids[(current + int(id)) % len(tab_ids)]
                else:
                    if int(id) >= len(tab_ids):
                        id = len(tab_ids) - 1

                    id = tab_ids[id]

        if id != self.default_tab:
            self.background.target.activate_target(id)
            self.default_tab = id

        return self.default

    def close_tab(self, id):
        success = self.background.target.close_target(id)
        self.sync()
        return success

    def execute_script(self, script, scope=None):
        """
        Execute an automation script against this browser session.

        Args:
            script (str):
                The full text of the script to execute.

            scope (`webfriend.scripting.Scope`, optional):
                An optional initial scope to execute the script from.  This can be used to provide the
                script with initial values for variables used in the script.
        """
        return execute_script(self, script, scope)

    def create_temp_profile(self, tempdir=None):
        self.temp_profile_path = tempfile.mkdtemp(
            prefix='webfriend-',
            suffix='-chrome',
            dir=tempdir
        )

        with open(os.path.join(self.temp_profile_path, 'First Run'), 'wb') as preferences:
            data = json.dumps(self.temp_profile_preferences, indent=4)
            preferences.write(data)
            self.args = [
                '--user-data-dir={}'.format(self.temp_profile_path)
            ] + self.args

        logging.info('Created temporary profile at {}'.format(self.temp_profile_path))
        return self.temp_profile_path

    def destroy_temp_profile(self):
        if self.temp_profile_path:
            shutil.rmtree(self.temp_profile_path)
            self.temp_profile_path = None
            return True

        return False



from __future__ import absolute_import
from __future__ import unicode_literals
from unittest import TestCase
from webfriend.scripting.scope import Scope


class ScopeTest(TestCase):
    def test_inheritance_and_overrides(self):
        top = Scope({
            'a': 1,
        })

        one_deep = Scope({
            'b': 2,
        }, parent=top)

        two_deep = Scope({
            'c': 3,
        }, parent=one_deep)

        # ensure that all values are as we expect them from various perspectives
        self.assertEqual(1, top['a'])
        self.assertTrue(top.is_local('a'))

        self.assertEqual(1, one_deep['a'])
        self.assertEqual(2, one_deep['b'])

        self.assertFalse(one_deep.is_local('a'))
        self.assertTrue(one_deep.is_local('b'))

        self.assertEqual(1, two_deep['a'])
        self.assertEqual(2, two_deep['b'])
        self.assertEqual(3, two_deep['c'])

        self.assertFalse(two_deep.is_local('a'))
        self.assertFalse(two_deep.is_local('b'))
        self.assertTrue(two_deep.is_local('c'))

        # ensure that setting a value on a child scope appropriately routes the set to the
        # correct owning scope, and that all children immediately see the update
        two_deep['a'] = 11
        self.assertFalse(two_deep.is_local('a'))
        self.assertEqual(11, two_deep['a'])

        self.assertFalse(one_deep.is_local('a'))
        self.assertEqual(11, one_deep['a'])

        self.assertTrue(top.is_local('a'))
        self.assertEqual(11, top['a'])

        # ensure that force-setting a value creates that value on the scope it was set on,
        # regardless of parent membership
        two_deep.set('b', 22, force=True)
        self.assertTrue(two_deep.is_local('b'))
        self.assertEqual(22, two_deep['b'])

        # make sure it's still the old value from the parent's perspective
        self.assertTrue(one_deep.is_local('b'))
        self.assertEqual(2, one_deep['b'])

        # and that top can't see it at all
        self.assertFalse(top.is_local('b'))
        self.assertIsNone(top.get('b'))

    def test_nested_inheritance_and_overrides(self):
        top = Scope({
            'a': {
                'thing': 1,
            },
        })

        one_deep = Scope({
            'b': {
                'thing': 2,
            },
        }, parent=top)

        two_deep = Scope({
            'c': {
                'thing': 3,
            },
        }, parent=one_deep)

        # ensure that all values are as we expect them from various perspectives
        self.assertEqual(1, top['a.thing'])
        self.assertTrue(top.is_local('a.thing'))

        self.assertEqual(1, one_deep['a.thing'])
        self.assertEqual(2, one_deep['b.thing'])

        self.assertFalse(one_deep.is_local('a.thing'))
        self.assertTrue(one_deep.is_local('b.thing'))

        self.assertEqual(1, two_deep['a.thing'])
        self.assertEqual(2, two_deep['b.thing'])
        self.assertEqual(3, two_deep['c.thing'])

        self.assertFalse(two_deep.is_local('a.thing'))
        self.assertFalse(two_deep.is_local('b.thing'))
        self.assertTrue(two_deep.is_local('c.thing'))

        # ensure that setting a value on a child scope appropriately routes the set to the
        # correct owning scope, and that all children immediately see the update
        two_deep['a.thing'] = 11
        self.assertFalse(two_deep.is_local('a.thing'))
        self.assertEqual(11, two_deep['a.thing'])

        self.assertFalse(one_deep.is_local('a.thing'))
        self.assertEqual(11, one_deep['a.thing'])

        self.assertTrue(top.is_local('a.thing'))
        self.assertEqual(11, top['a.thing'])

        # ensure that force-setting a value creates that value on the scope it was set on,
        # regardless of parent membership
        two_deep.set('b.thing', 22, force=True)
        self.assertTrue(two_deep.is_local('b.thing'))
        self.assertEqual(22, two_deep['b.thing'])

        # make sure it's still the old value from the parent's perspective
        self.assertTrue(one_deep.is_local('b.thing'))
        self.assertEqual(2, one_deep['b.thing'])

        # and that top can't see it at all
        self.assertFalse(top.is_local('b.thing'))
        self.assertIsNone(top.get('b.thing'))

        # I'll just leave this one here...
        a = one_deep

        # assert that this remains true
        a['g.thing'] = None
        self.assertIsNone(a['g.thing'])

        # ensure that creating a value on one_deep is visible but not local to two_deep, and
        # not visible from top
        one_deep['z.y.x'] = 42
        self.assertTrue(one_deep.is_local('z'))
        self.assertTrue(one_deep.is_local('z.y'))
        self.assertTrue(one_deep.is_local('z.y.x'))
        self.assertEqual(42, one_deep['z.y.x'])

        self.assertFalse(two_deep.is_local('z'))
        self.assertFalse(two_deep.is_local('z.y'))
        self.assertFalse(two_deep.is_local('z.y.x'))
        self.assertEqual(42, two_deep['z.y.x'])

    def test_scope_nested_retrieval(self):
        scope = Scope({
            'top': True,
            'nested': {
                'values': {
                    'are': [
                        'fun',
                    ]
                }
            }
        })

        self.assertEqual(True, scope.get('top'))

        self.assertEqual({
            'values': {
                'are': ['fun'],
            }
        }, scope.get('nested'))

        self.assertEqual({
            'are': ['fun'],
        }, scope.get('nested.values'))

        self.assertEqual(['fun'], scope.get('nested.values.are'))

        self.assertIsNone(scope.get('nested.nonexistent'))

        self.assertIsNone(scope.get('nested.values.nonexistent'))

        self.assertIsNone(scope.get('nested.double.nonexistent'))

        self.assertTrue('top' in scope)
        self.assertTrue('nested' in scope)
        self.assertTrue('nested.values' in scope)
        self.assertTrue('nested.values.are' in scope)
        self.assertFalse('nested.values.nonexistent' in scope)
        self.assertFalse('nested.nonexistent.are' in scope)
        self.assertFalse('nested.double.nonexistent' in scope)

        scope.unset('nested.values')
        self.assertFalse('nested.values.are' in scope)
        self.assertFalse('nested.values' in scope)
        self.assertTrue('nested' in scope)



from __future__ import absolute_import
from __future__ import unicode_literals
from urlparse import urlparse
from uuid import uuid4
from webfriend import exceptions
from webfriend import rpc, utils
from webfriend.scripting.commands.base import CommandProxy
from webfriend.scripting.execute import execute_script
from webfriend.scripting.environment import Environment
from webfriend.scripting.scope import Scope
from webfriend.scripting.parser.exceptions import UserError
from webfriend.utils import autotype
from webfriend.utils.docs import document_commands
import json
import logging
import os
import re
import time
import urlnorm


class CoreProxy(CommandProxy):
    """
    These represent very common tasks that one is likely to perform in a browser, such as
    navigating to URLs, filling in form fields, and performing input with the mouse and keyboard.
    """

    @classmethod
    def qualify(cls, name):
        return name

    def configure(
        self,
        events=None,
        demo=None,
        user_agent=None,
        extra_headers=None,
        cache=None,
        console=None
    ):
        """
        Configures various features of the Remote Debugging protocol and provides environment
        setup.

        #### Arguments

        - **events** (`list`, optional):

            A list of strings specifying what kinds of events Chrome should send to this
            client.  Valid values are: `console`, `dom`, `network`, `page`.

        - **demo** (`dict`, optional):

            A section describing various runtime options useful for demonstrations and
            walkthroughs.

            - **delay** (`int`, optional):

                If specified, scripts will sleep for this amount of time (in milliseconds) between
                each command that is processed.  This is useful for slowing down the visual
                effects of commands to a rate that is easier to see.

        - **user_agent** (`str`, optional):

            If specified, this will be the User-Agent header value that is sent with all HTTP(S)
            requests initiated from here on.

        - **extra_headers** (`dict`, optional):

            If specified, these headers will be included in all HTTP(S) requests initiated from
            here on.  An empty dict will clear previously set headers.

        - **cache** (`bool`, optional):

            Whether caching is enabled or not for this session.

        - **console** (`bool`, optional):

            Whether console messages emitted from pages are logged to standard error.
        """
        if events and hasattr(events, 'values') and isinstance(events.values, list):
            for domain in events.values:
                domain = str(domain).lower()

                if hasattr(self.tab, domain):
                    r = getattr(self.tab, domain)

                    if isinstance(r, rpc.Base):
                        logging.info('Enabling events for domain "{}"'.format(domain))
                        r.enable()

        if isinstance(demo, dict):
            if 'delay' in demo:
                self.environment.set_execution_option(
                    'demo.post_command_delay',
                    float(demo['delay'])
                )

        if isinstance(user_agent, basestring):
            self.tab.network.set_user_agent(user_agent)

        if isinstance(extra_headers, dict):
            self.tab.network.set_headers(extra_headers)

        if cache is True:
            self.tab.network.enable_cache()
        elif cache is False:
            self.tab.network.disable_cache()

        if console is True:
            self.tab.enable_console_messages()
        else:
            self.tab.disable_console_messages()

    def go(
        self,
        uri,
        referrer='random',
        wait_for_load=True,
        timeout=30000,
        clear_requests=True,
        continue_on_error=False
    ):
        """
        Nagivate to a URL.

        #### Arguments

        - **referrer** (`str`, optional):

            If a URL is specified, it will be used as the HTTP Referer [sic] header field when
            going to the given page. If the URL of the currently-loaded page and the referrer
            are the same, the page will no change.

            For this reason, you may specify the special value 'random', which will generate a
            URL with a randomly generated path component to ensure that it is always different
            from the current page. Specifying None will omit the field from the request.

        - **wait_for_load** (`bool`):

            Whether to block until the page has finished loading.

        - **timeout** (`int`):

            The amount of time, in milliseconds, to wait for the the to load.

        - **clear_requests** (`bool`, optional):

            Whether the resources stack that is queried in [page::resources](#pageresources) and
            [page::resource](#pageresource) is cleared before navigating.  Set this to _false_ to
            preserve the ability to retrieve data that was loaded on previous pages.

        #### Returns
        The URL that was loaded (`str`)
        """

        if referrer is 'random':
            referrer = 'http://example.com/{}'.format(uuid4())

        if clear_requests:
            # since we've explicitly navigating, clear the network requests
            self.tab.dom.clear_requests()

        uri_p = urlparse(uri)

        if not len(uri_p.scheme):
            uri = 'https://{}'.format(uri)

        uri = urlnorm.norm(uri)

        reply = self.tab.page.navigate(uri, referrer=referrer)

        if wait_for_load:
            try:
                self.tab.wait_for('Page.loadEventFired', timeout=timeout)
            except exceptions.TimeoutError:
                if continue_on_error:
                    logging.error('Timed out waiting for page load event.')
                else:
                    raise

        return reply

    def reload(self):
        self.tab.page.reload()

    def stop(self):
        self.tab.page.stop()

    def wait(self, milliseconds):
        """
        Pauses execution of the current script for the given number of milliseconds.

        #### Arguments

        - **milliseconds** (`int`):

            The number of milliseconds to sleep for; can be fractional.

        #### Returns
        The number of milliseconds.
        """

        time.sleep(milliseconds / 1e3)
        return milliseconds

    def resize(
        self,
        width=0,
        height=0,
        scale=0,
        mobile=False,
        fit_window=False,
        orientation=None,
        angle=0
    ):
        """
        Resizes the active viewport of the current page using the Chrome Device Emulation API. This
        does not resize the window itself, but rather the area the current page interprets the
        window to be.

        This is useful for setting the size of the area that will be rendered for screenshots and
        screencasts.

        #### Arguments

        - **width** (`int`, optional):
            The desired width of the viewport.

        - **height** (`int`, optional):
            The desired height of the viewport.

        - **scale** (`float`, optional):
            The scaling factor of the content.

        - **mobile** (`bool`, dict, optional):
            Whether to emulate a mobile device or not.  If a dict is provided, mobile emulation
            will be enabled and configured using the following keys:

            - *width* (`int`, optional):
                The width of the mobile screen to emulate.

            - *height* (`int`, optional):
                The height of the mobile screen to emulate.

            - *x* (`int`, optional):
                The horizontal position of the currently viewable portion of the mobile screen.

            - *y* (`int`, optional):
                The vertical position of the currently viewable portion of the mobile screen.

        - **fit_window** (`bool`, optional):
            Whether to fit the viewport contents to the available area or not.

        - **orientation** (`str`, optional):
            Which screen orientation to emulate, if any. Can be one of: `portraitPrimary`,
            `portraitSecondary`, `landscapePrimary`, `landscapeSecondary`.

        - **angle** (`int`, optional):
            The angle of the screen to emulate (in degrees; 0-360).

        #### Returns
        A `dict` containing the resulting *width* and *height* as keys.
        """

        if hasattr(mobile, 'as_dict'):
            _mobile   = True
            cfg       = mobile.as_dict()
            mobile_sw = cfg.get('width', width)
            mobile_sh = cfg.get('height', height)
            mobile_x  = cfg.get('x', 0)
            mobile_y  = cfg.get('y', 0)
        else:
            _mobile   = False
            mobile_sw = 0
            mobile_sh = 0
            mobile_x  = 0
            mobile_y  = 0

        self.tab.emulation.set_device_metrics_override(
            width=width,
            height=height,
            device_scale_factor=scale,
            mobile=_mobile,
            fit_window=fit_window,
            mobile_screen_width=mobile_sw,
            mobile_screen_height=mobile_sh,
            mobile_position_x=mobile_x,
            mobile_position_y=mobile_y,
            screen_orientation_type=orientation,
            screen_orientation_angle=angle,
        )

        return {
            'width':  width,
            'height': height,
        }

    def put(self, *args, **kwargs):
        """
        Store a value in the current scope.  Strings will be automatically converted into the
        appropriate data types (float, int, bool) if possible.

        #### Arguments

        If a single argument is given, automatic type detection will be applied to it and the
        resulting value will be returned.

        If options are provided, they will be interpreted as an object, each of whose values will
        have automatic type detection applied.  The resulting object will be returned.

        #### Returns
        The given value with automatic type detection applied.
        """
        if len(args) == 1:
            return utils.autotype(args[0])

        elif len(args) > 1:
            return [utils.autotype(a) for a in args]

        elif len(kwargs):
            return dict([
                (k, utils.autotype(v)) for k, v in kwargs.items()
            ])

        return None

    def log(self, line=None, level='info', indent=4, **kwargs):
        """
        Outputs a line to the log.

        #### Arguments

        - **line** (`str`):
                A line of text that will have all current variables, as well as any given
                kwargs, interpolated into it using the Python format() function.

        - **level** (`str`):
                The logging severity level to out as. Can be one of: 'debug', 'info', 'warning',
                or 'error'.

        - **indent** (`int`):
                If 'line' is a dictionary, list, or tuple, it will be printed as a JSON document
                with an indentation of this many spaces per level.  The special value -1 will
                disable JSON serialization for these types.

        - **kwargs**:
                All remaining arguments will be passed along to format() when interpolating 'line'.

        #### Returns
        None

        #### Raises
        `AttributeError` if the specified log level is not known.
        """
        # handle the case where we want to log a data structure without options or a format string
        if line is None:
            line = kwargs

        if hasattr(self.environment.log, level):
            if isinstance(line, (dict, list, tuple)) and indent >= 0:
                try:
                    line = json.dumps(line, indent=4)
                except:
                    pass

            # actually log the line
            getattr(self.environment.log, level)(line)
            return None
        else:
            raise AttributeError("Unknown log level '{}'".format(level))

    def fail(self, message):
        """
        Immediately exit the script in an error-like fashion with a specific message.

        #### Arguments

        - **message** (`str`):

            The message to display whilst exiting immediately.

        #### Raises
        - `webfriend.exceptions.UserError`
        """
        self.environment.log.error(message)
        raise UserError(message)

    def rpc(self, method, **kwargs):
        """
        Directly call an RPC method with the given parameters.

        #### Arguments

        - **method** (`str`):

        The name of the backend RPC method to call.

        - **kwargs**:

        Zero of more arguments to pass in the 'params' section of the RPC call.

        #### Returns
        A `dict` representation of the `webfriend.rpc.Reply` class.
        """
        return self.tab.rpc(method, **kwargs).as_dict()

    def wait_for(self, event_name, timeout=30000, match=None):
        """
        Block until a specific event is received, or until **timeout** elapses (whichever comes
        first).

        #### Arguments

        - **event_name** (`str`):

            The name of the event to wait for.

        - **timeout** (`int`):

            The timeout, in milliseconds, before raising a `webfriend.exceptions.TimeoutError`.

        - **match** (`dict`, optional):

            If specified, all keys in the given object must correspond to keys in the received
            event payload, and the values must match.  Regular expressions must match the
            corresponding payload value, and all other types must match exactly.

        #### Returns
        `webfriend.rpc.Event`

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """
        if isinstance(match, dict):
            started_at = time.time()
            eventstream = self.tab.wait_for_caller_response(event_name, timeout=timeout)

            for event in eventstream:
                if event.matches_criteria(match):
                    try:
                        eventstream.send(True)
                    except StopIteration:
                        pass

                    return {
                        'sequence': [event],
                        'duration': (time.time() - started_at),
                    }
        else:
            return self.tab.wait_for(event_name, timeout=timeout)

    def wait_for_idle(self, idle, events=[], timeout=30000, poll_interval=250):
        """
        Blocks for a specified amount of time _after_ an event has been received, or until
        **timeout** elapses (whichever comes first).

        This is useful for waiting for events to occur after performing an action, then giving some
        amount of time for those events to "settle" (e.g.: allowing the page time to react to those
        events without knowing ahead of time what, if any, listeners will be responding.)  A common
        use case for this would be to wait a few seconds _after_ a resize has occurred for anything
        that just loaded to finish doing so.


        #### Arguments

        - **idle** (`int`):

            The amount of time, in milliseconds, that the event stream should be idle before
            returning.

        - **events** (`list`, optional):

            If not empty, the **idle** time will be interpreted as the amount of time since _any
            of these specific events_ have occurred.  The default is to wait for the browser to be
            idle with respect to _any_ events.

        - **timeout** (`int`):

            The maximum amount of time to wait before raising a
            `webfriend.exceptions.TimeoutError`.

        - **poll_interval** (`int`):

            How often to check the event timings to see if the idle time has elapsed.

        #### Returns
        An `int` representing the number of milliseconds we waited for.

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """
        if hasattr(events, 'values'):
            events = events.values

        return self.tab.wait_for_idle(
            idle,
            events=events,
            timeout=timeout,
            poll_interval=poll_interval
        )

    def wait_for_load(self, timeout=30000, idle_time=500):
        """
        Blocks until the "Page.loadEventFired" event has fired, or until timeout elapses (whichever
        comes first).

        #### Arguments

        - **timeout** (`int`):

            The timeout, in milliseconds, before raising a `webfriend.exceptions.TimeoutError`.

        #### Returns
        `webfriend.rpc.Event`

        #### Raises
        `webfriend.exceptions.TimeoutError`
        """

        if idle_time:
            return self.tab.wait_for_idle(idle_time, events=[
                'Page.loadEventFired',
            ], timeout=timeout)
        else:
            return self.tab.wait_for('Page.loadEventFired', timeout=timeout)

    def type(self, text, **kwargs):
        """
        See: `webfriend.rpc.Input.type_text`
        """
        return self.tab.input.type_text(text, **kwargs)

    def focus(self, selector):
        """
        Focuses the given HTML element described by **selector**.  One and only one element may
        match the selector.

        #### Arguments

        - **selector** (`str`):

            The page element to focus, given as a CSS-style selector, an ID (e.g. "#myid"), or an
            XPath query (e.g.: "xpath://body/p").

        #### Returns
        The matching `webfriend.rpc.DOMElement` that was given focus.

        #### Raises
        - `webfriend.exceptions.EmptyResult` if zero elements were matched, or
        - `webfriend.exceptions.TooManyResults` if more than one elements were matched.
        """
        elements = self.tab.dom.query_all(selector)
        self.tab.dom.ensure_unique_element(selector, elements)
        element = elements['nodes'][0]

        self.tab.dom.focus(element.id)
        return element

    def click(self, selector=None, x=None, y=None, unique_match=True, **kwargs):
        """
        Click on HTML element(s) or on a specific part of the page.  More complex click operations
        are supported (e.g.: double clicking, drag and drop) by supplying **x**/**y** coordinates
        directly.

        #### Arguments

        - **selector** (`str`, optional):

            The page element to focus, given as a CSS-style selector, an ID (e.g. "#myid"), or an
            XPath query (e.g.: "xpath://body/p").

        - **x** (`int`, optional):

            If **selector** is not specified, this is the X-coordinate component of the location
            to click at.

        - **y** (`int`, optional):

            If **selector** is not specified, this is the Y-coordinate component of the location
            to click at.

        - **unique_match** (`bool`):

            For **selector** matches, whether there can be one and only one match to click on. If
            false, every matched element will be clicked on in the order they were matched in.

        - **kwargs**:

            Only applies to **x**/**y** click events, see: `webfriend.rpc.Input.click_at`.

        #### Returns
        A `list` of elements that were clicked on.

        #### Raises
        For **selector**-based events:

        - `webfriend.exceptions.EmptyResult` if zero elements were matched, or
        - `webfriend.exceptions.TooManyResults` if more than one elements were matched.
        """
        if selector:
            elements = self.tab.dom.select_nodes(selector)
            results = []

            if unique_match:
                self.tab.dom.ensure_unique_element(selector, elements)

            for element in elements['nodes']:
                results.append(element.click())

            return results

        elif x is None or y is None:
            raise ValueError("Either 'selector' or 'x' and 'y' must be specified")

            return [
                self.tab.input.click_at(x, y, **kwargs)
            ]

    def field(self, selector, value, autoclear=True):
        """
        Locate and enter data into a form input field.

        #### Arguments

        - **selector** (`str`):

            The page element to enter data into, given as a CSS-style selector, an ID
            (e.g. "#myid"), or an XPath query (e.g.: "xpath://body/p").

        - **value** (`str`):

            The text value to type into the located field.

        - **autoclear** (`bool`, optional):

            Whether to clear the existing contents of the field before entering new data.

        #### Returns
        The text that was entered, as a string.
        """
        if not isinstance(value, basestring):
            raise ValueError("'value' must be specified")

        elements = self.tab.dom.select_nodes(selector)
        field = self.tab.dom.ensure_unique_element(selector, elements)

        if autoclear:
            field['value'] = ''

        field.focus()
        return self.type(value)

    def scroll_to(self, selector=None, x=None, y=None):
        """
        Scroll the viewport to the given location, either that of the named element or, if
        provided, the specfic (X,Y) coordinates relative to the top-left of the current page.

        #### Arguments

        - **selector** (`str`, optional):

            The page element to scroll to, given as a CSS-style selector, an ID (e.g. "#myid"), or
            an XPath query (e.g.: "xpath://body/p").

        - **x**, **y** (`int`, optional):

            If both **x* and **y** are provided, these are the coordinates to scroll to.

        #### Returns
        The result of the scroll operation.
        """
        if selector:
            elements = self.tab.dom.select_nodes(selector)
            self.tab.dom.ensure_unique_element(selector, elements)

            return elements['nodes'][0].scroll_to()

        elif x is None or y is None:
            raise ValueError("Either 'selector' or 'x' and 'y' must be specified")
            return self.tab.dom.root.scroll_to(x, y)

    def select(self, *args, **kwargs):
        """
        See: `webfriend.rpc.DOM.select_nodes`
        """
        return self.tab.dom.select_nodes(*args, **kwargs)

    def xpath(self, *args, **kwargs):
        """
        See: `webfriend.rpc.DOM.xpath`
        """
        return self.tab.dom.xpath(*args, **kwargs)

    def switch_root(self, selector=None):
        """
        Change the current selector scope to be rooted at the given element.
        """
        if selector is None:
            return self.tab.dom.root_to_top()
        else:
            return self.tab.dom.root_to(selector)

    def tabs(self, sync=True):
        """
        Return a description of all of the currently-open tabs.

        #### Arguments

        - **sync** (`bool`):

            Whether to perform a preemptive sync with the browser before returning the tab
            descriptions.

        #### Returns
        A `list` of `dicts` describing all browser tabs currently open.  Each `dict` will at least
        contain the keys:

        - *id* (`str`):

            The tab ID that can be used with other tab management commands.

        - *url* (`str`):

            The URL of the tab being described.

        - *webSocketDebuggerUrl* (`str`):

            The URL of the inspection Websocket used to issue and receive RPC traffic.

        - *target* (`bool`):

            Whether the tab being described is the active tab that other commands will be issued
            against.
        """
        if sync:
            self.browser.sync()

        return [
            t.as_dict() for t in self.browser.tabs.values()
        ]

    def new_tab(self, url, width=None, height=None, autoswitch=True):
        """
        Open a new tab and navigate to the given URL.

        #### Arguments

        - **url** (`str`):

            The URL that the new tab will be navigated to.

        - **width**, **height** (`int`, optional):

            If provided, these represent the width and height (in pixels) that the new tab should
            be created with.

        - **autoswitch** (`bool`, optional):

            Whether to automatically switch to the newly-created tab as the active tab for
            subsequent commands.

        #### Returns
        A `str` representing the ID of the newly-created tab.
        """
        tab_id = self.browser.create_tab(url, width=width, height=height)

        if autoswitch:
            self.browser.switch_to_tab(tab_id)

        return tab_id

    def switch_tab(self, tab_id):
        """
        See: `webfriend.browser.Chrome.switch_to_tab`
        """
        self.browser.switch_to_tab(tab_id)
        return self.tabs(sync=False)

    def close_tab(self, tab_id=None):
        """
        Close the tab identified by the given ID.

        #### Arguments

        - **tab_id** (`str`):

            The ID of the tab to close.

        #### Returns
        A `bool` value representing whether the tab was closed successfully or not.
        """
        if not tab_id:
            tab_id = self.browser.default_tab

        return self.browser.close_tab(tab_id)

    def javascript(self, body=None, file=None, expose_variables=True):
        """
        Inject Javascript into the current page, evaluate it, and return the results.  The script
        is wrapped in an anonymous function whose return value will be returned from this command
        as a native data type.

        By default, scripts will have access to all local variables in the calling script that are
        defined at the time of invocation.  They are available to injected scripts as a plain
        object accessible using the `this` variable.

        #### Arguments

        - **body** (`str`, optional):

            A string value that represents the script to be injected and executed.

        - **file** (`str`, optional):

            A filename that will loaded and injected into the browser.

        - **expose_variables** (`bool`):

            Whether to expose all local variables to the injected script or not.

        #### Returns
        Whatever data was returned from the injected script using a `return` statement,
        automatically parsed into native data types.  Objects, arrays, and all scalar types are
        supported as return values.
        """
        if not body and not file:
            raise ValueError("Must specify either body or file")

        if file:
            body = open(file, 'r').read()

        if expose_variables:
            data = self.scope.as_dict()
        else:
            data = {}

        return self.tab.evaluate(body, data=data, calling_context=self.environment)

    def env(self, name, fallback=None, ignore_empty=True, detect_type=True, joiner=None):
        """
        Retrieves a system environment variable and returns the value of it, or a fallback value if
        the variable does not exist or (optionally) is empty.

        #### Arguments

        - **name** (`str`):

            The name of the environment variable.  Matches are case-insensitive, and the last
            variable to be defined for a given key is the value that will be returned.

        - **fallback** (any):

            The value to return if the environment variable does not exist, or (optionally) is
            empty.

        - **ignore_empty** (`bool`):

            Whether empty values should be ignored or not.

        - **detect_type** (`bool`):

            Whether automatic type detection should be performed or not.

        - **joiner** (`str`, optional):

            If specified, this string will be used to split matching values into a list of values.
            This is useful for environment variables that contain multiple values joined by a
            separator (e.g: the `PATH` variable.)

        #### Returns
        The value of the environment variable **name**, or a list of values if **joiner** was
        specified. If **name** is non-existent or was empty, **fallback** will be returned instead.
        """
        value = None

        # perform case-insensitive search of all environment variables
        for k, v in os.environ.items():
            if k.upper() == name.upper():
                value = v
                break

        if value is None:
            return fallback

        # trim whitespace
        value = value.strip()

        if isinstance(joiner, basestring):
            value = value.split(joiner)

        # handle empty values AND empty lists post-split
        if ignore_empty and not len(value):
            return fallback

        # perform type detection (if specified)
        if detect_type:
            if isinstance(value, list):
                value = [autotype(v) for v in value]
            else:
                value = autotype(value)

        return value

    def require(self, plugin_name, package_format='webfriend.scripting.commands.{}'):
        """
        Loads a named plugin into the current environment.

        #### Arguments

        - **plugin_name** (`str`):

            The name of the plugin to load.  This corresponds to the name of a Python module that
            contains subclasses of `webfriend.scripting.commands.base.CommandProxy`.

        - **package_format** (`str`):

            Specifies which Python package contains the module named in **plugin_name**.  The
            default is to assume plugins are built as namespaced modules that overlay the core
            import tree at `webfriend.scripting.commands.<plugin_name>`.

        #### Returns
        The value of **plugin_name** if the load was successful.
        """
        self.environment.register_by_module_name(plugin_name)
        return plugin_name

    def run(
        self,
        script_name,
        data=None,
        isolated=True,
        preserve_state=True,
        merge_scopes=False,
        result_key=Environment.default_result_key
    ):
        """
        Evaluates another Friendscript loaded from another file.

        #### Arguments

        - **script_name** (`str`):

            The filename or basename of the file to search for in the `WEBFRIEND_PATH` environment
            variable to load and evaluate.  The `WEBFRIEND_PATH` variable behaves like the the
            traditional *nix `PATH` variable, wherein multiple paths can be specified as a
            colon-separated (`:`) list.  The current working directory will always be checked
            first.

        - **data** (`dict`, optional):

            If specified, these values will be made available to the evaluated script before it
            begins execution.

        - **isolated** (`bool`):

            Whether the script should have access to the calling script's variables or not.

        - **preserve_state** (`bool`):

            Whether event handlers created in the evaluated script should remain defined after the
            script has completed.

        - **merge_scopes** (`bool`):

            Whether the scope state at the end of the script's evaluation should be merged into the
            current execution scope.  Setting this to true allows variables defined inside of the
            evaluated script to stay defined after the script has completed.  Otherwise, only the
            value of the **result_key** variable is returned as the result of this command.

        - **result_key** (`str`):

            Defines the name of the variable that will be read from the evaluated script's scope
            and returned from this command.  Defaults to "result", which is the same behavior as
            all other commands.

        #### Returns
        The value of the variable named by **result_key** at the end of the evaluated script's
        execution.

        #### Raises
        Any exception that can be raised from Friendscript.
        """
        script = None
        final_script_name = None
        path_prefixes = ['.']

        # makes the ".fs" optional when passing scripts as arguments
        script_name = re.sub(r'\.fs$', '', script_name) + '.fs'

        # setup scopes
        if isolated:
            scope = Scope()
        else:
            scope = Scope(parent=self.scope)

        # if data is specified, set these values in the evaluated script's scope
        if isinstance(isolated, dict):
            scope.update(isolated)

        # process WEBFRIEND_PATH envvar
        for prefix in os.environ.get('WEBFRIEND_PATH', '').split(':'):
            path_prefixes.append(prefix)

        # search for file in all prefixes
        for prefix in path_prefixes:
            s = os.path.join(prefix, script_name)

            if os.path.isfile(s):
                final_script_name = s
                script = open(final_script_name, 'r').read()
                break

        # validate the script was read
        if script is None:
            raise exceptions.NotFound('Unable to locate script "{}" in any path'.format(
                script_name
            ))

        elif isinstance(script, str):
            script = script.decode('UTF-8')

        # evaluate the script
        logging.debug('Evaluating script {}'.format(final_script_name))
        scope = execute_script(
            self.browser,
            script,
            scope=scope,
            preserve_state=preserve_state
        )

        # if not using an isolated scope, then the top-level keys that were modified in this script
        # are set in our current scope (as if the included code ran inline)
        if merge_scopes:
            logging.debug(
                'Updating {} variables in calling scope with result scope (iso={})'.format(
                    len(scope),
                    isolated
                )
            )

            self.scope.update(scope)

        return scope.get('result')

    def help(self, command):
        print('\n'.join(document_commands(
            oneshot=True,
            commands=[command],
        )).rstrip('\n'))



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting import parser
from webfriend.scripting.environment import Environment, IgnoreResults
from webfriend.scripting.scope import Scope
from webfriend.scripting.commands import *  # noqa
from threading import Thread
import logging
import time


def execute_script(browser, script, scope=None, environment=None, preserve_state=False):
    # setup the environment
    if not environment:
        environment = Environment(scope, browser=browser)

    if not scope:
        scope = environment.scope

    callbacks = set()

    # load and parse the script
    friendscript = parser.Friendscript(data=script, environment=environment)

    # tell the environment about the calling script
    environment.script = friendscript

    if browser:
        # setup event handlers for this execution
        for handler in friendscript.handlers:
            callback_id = browser.default.on(
                parser.to_value(handler.pattern, scope),
                _handle_event(friendscript, handler, scope)
            )

            logging.debug('Add event handler {}'.format(callback_id))
            callbacks.add(callback_id)

    # Script Execution starts NOW
    # ---------------------------------------------------------------------------------------------
    try:
        # recursively evaluate all blocks and nested blocks starting from the top level
        for block in friendscript.blocks:
            try:
                evaluate_block(friendscript, block, scope)
            except parser.exceptions.ScriptError:
                raise
            except Exception as e:
                logging.debug('Exception')
                raise
                raise parser.exceptions.ScriptError('{}'.format(e), model=block)

    finally:
        if not preserve_state:
            # unregister the event handlers we created for this run
            for callback_id in callbacks:
                logging.debug('Remove event handler {}'.format(callback_id))
                browser.default.remove_handler(callback_id)

    # ...and done.
    # ---------------------------------------------------------------------------------------------

    # return the final state after script execution
    return scope


def evaluate_block(scriptmgr, block, scope):
    environment = scriptmgr.environment
    # line, column = scriptmgr.get_item_position(block)
    # print('{} = line {}, column: {}'.format(block.__class__, line, column))

    # Assignment
    # ---------------------------------------------------------------------------------------------
    if isinstance(block, parser.variables.Assignment):
        block.assign(environment, scope)

    # Directives
    # ---------------------------------------------------------------------------------------------
    elif isinstance(block, parser.lang.Directive):
        if block.is_unset:
            for var in block.variables:
                scope.unset(var.as_key(scope))

    # Expressions
    # ---------------------------------------------------------------------------------------------
    elif isinstance(block, parser.lang.Expression):
        result, put_result_in = block.process(environment, scope)

        if put_result_in:
            scope.set(put_result_in.as_key(scope), result)

    # Commands
    # ---------------------------------------------------------------------------------------------
    elif isinstance(block, parser.commands.CommandSequence):
        # for each command in the pipeline...
        for command in block.commands:
            key, value = environment.execute(command, scope)

            if not isinstance(value, IgnoreResults):
                value = parser.to_value(value, scope)

                if key is not None:
                    scope.set(key, value)

            # perform delay if we have one
            if environment.has_execution_option('demo.post_command_delay'):
                delay = environment.get_execution_option('demo.post_command_delay')
                time.sleep(delay / 1e3)

    # If / Else If / Else
    # ---------------------------------------------------------------------------------------------
    elif isinstance(block, parser.conditions.IfElseBlock):
        subscope = Scope(parent=scope)

        for subblock in block.get_blocks(environment, scope=subscope):
            evaluate_block(scriptmgr, subblock, subscope)

    # Loops
    # ---------------------------------------------------------------------------------------------
    elif isinstance(block, parser.loops.LoopBlock):
        loopscope = Scope(parent=scope)

        for i, _, subblock, scope in block.execute_loop(environment, scope=loopscope):
            try:
                evaluate_block(scriptmgr, subblock, scope=scope)

            except parser.loops.FlowControlMultiLevel as e:
                if e.levels > 1:
                    e.levels -= 1
                    raise
                elif isinstance(e, parser.loops.FlowControlBreak):
                    break
                elif isinstance(e, parser.loops.FlowControlContinue):
                    continue

    # Flow Control
    # ---------------------------------------------------------------------------------------------
    elif isinstance(block, parser.loops.FlowControlWord):
        if block.is_break:
            raise parser.loops.FlowControlBreak('break', levels=block.levels)

        elif block.is_continue:
            raise parser.loops.FlowControlContinue('continue', levels=block.levels)

        else:
            raise parser.exceptions.ScriptError('Unrecognized statement')


def _handle_event(scriptmgr, handler, scope):
    def handle(e):
        def actual(event):
            local_scope_data = {
                'event': {
                    'name':     event.event,
                    'instance': event,
                    'data':     event.payload,
                },
            }

            if handler.isolated:
                local_scope = Scope(data=local_scope_data)
            else:
                local_scope = Scope(data=local_scope_data, parent=scope)

            # execute all commands, accumulating output into the local state
            for block in handler.blocks:
                try:
                    evaluate_block(scriptmgr, block, local_scope)
                except parser.exceptions.ScriptError:
                    raise
                except Exception as e:
                    logging.debug('Exception')
                    raise parser.exceptions.ScriptError(str(e), model=block)

            return local_scope

        Thread(target=actual, args=(e,)).start()

    return handle



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.commands.base import CommandProxy
from webfriend import exceptions
import netrc
import os
import stat


class UtilsProxy(CommandProxy):
    qualifier = 'utils'
    doc_name = 'Utilities'

    def netrc(self, hostname, filename='~/.netrc'):
        """
        Retrieves a username and password stored in a `.netrc`-formatted file.

        #### Arguments

        - **value** (`str`):

            The value to apply the operation to.

        #### Returns
        A 3-element tuple representing the hostname, username, and password of the matching record.
        """
        filename = os.path.expanduser(filename)

        # verify the permissions on the netrc file
        if os.path.isfile(filename):
            netrc_stat = os.stat(filename)
            mode = netrc_stat.st_mode

            for perm, lbl in [
                (stat.S_IRGRP, 'group-readable'),
                (stat.S_IRGRP, 'group-writable'),
                (stat.S_IXGRP, 'group-executable'),
                (stat.S_IROTH, 'world-readable'),
                (stat.S_IROTH, 'world-writable'),
                (stat.S_IXOTH, 'world-executable'),
            ]:
                if mode & perm:
                    raise OSError("Refusing to read credentials from {} file".format(lbl))
        else:
            raise IOError("Cannot locate netrc file at {}".format(filename))

        # python, why would you return these in this order?
        auth = netrc.netrc(filename).authenticators(hostname)

        if auth is None:
            raise exceptions.NotFound("No credentials found for host '{}'".format(hostname))

        # seriously, why?
        # username, hostname, then password? i know this is a nit, but come on...
        # does everything in this language have to be crooked and weird?
        return auth[1], auth[0], auth[2]



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.commands.base import CommandProxy
import tempfile
import shutil
import os


class FileProxy(CommandProxy):
    def open(self, filename, read=True, write=True, truncate=False, append=False, binary=False):
        mode = ''

        if read:
            if write:
                if truncate:
                    mode = 'w+'
                else:
                    mode = 'r+'
            elif append:
                mode = 'a+'
            else:
                mode = 'r'
        elif write:
            if truncate:
                mode = 'w'
        elif append:
            mode = 'a'

        if not len(mode):
            raise AttributeError("Must specify at least one of: 'read', 'write', or 'append'")

        if binary:
            mode += 'b'

        return open(filename, mode)

    def close(self, handle):
        handle.close()

    def read(self, filename):
        return open(filename, 'r').read()

    def write(self, filename, data):
        if data:
            if isinstance(data, unicode):
                data = data.encode('UTF-8')

            if isinstance(data, str):
                open(filename, 'wb').write(data)
            else:
                shutil.copyfileobj(data, open(filename, 'wb'))

            return data

        return None

    def append(self, filename, data):
        if data:
            if isinstance(data, unicode):
                data = data.encode('UTF-8')

            if isinstance(data, str):
                with open(filename, 'a') as f:
                    f.write(data)
            else:
                shutil.copyfileobj(data, open(filename, 'a'))

    def temp(self, directory='.', prefix='webfriend-', suffix=''):
        tmp = tempfile.mkstemp(
            suffix=suffix,
            prefix=prefix,
            dir=directory,
        )

        return {
            'handle':   tmp[0],
            'filename': tmp[1],
        }

    def dirname(self, filename, trim_query_string=True):
        if trim_query_string:
            filename = filename.split('?', 1)[0]

        return os.path.dirname(filename)

    def basename(self, filename, trim_query_string=True):
        if trim_query_string:
            filename = filename.split('?', 1)[0]

        return os.path.basename(filename)

    def exists(self, filename, trim_query_string=True):
        if trim_query_string:
            filename = filename.split('?', 1)[0]

        return os.path.exists(filename)

    def mkdir(self, path, recursive=True, mode=0o755):
        if not os.path.exists(path):
            if recursive:
                os.makedirs(path, mode)
            else:
                os.makedir(path, mode)

        elif not os.path.isdir(path):
            raise OSError("Path '{}' exists, but is not a directory".format(path))

        return True



from __future__ import absolute_import
from __future__ import unicode_literals
from unittest import TestCase
from webfriend.scripting.execute import execute_script


class FormatProxyTest(TestCase):
    def setUp(self):
        self.maxDiff = 10000

    def _eval(self, script):
        return execute_script(None, script).as_dict()

    def test_autotype(self):
        self.assertEqual({
            'a': 42,
            'b': 3.1415,
            'c': True,
            'd': True,
            'e': True,
            'f': False,
            'g': False,
            'h': False,
            'i': -42,
            'j': -3.1415,
            'k': 0,
            'l': 1,
        }, self._eval("""
            fmt::autotype "42" -> $a
            fmt::autotype "3.1415" -> $b
            fmt::autotype "true" -> $c
            fmt::autotype "True" -> $d
            fmt::autotype "TRUE" -> $e
            fmt::autotype "false" -> $f
            fmt::autotype "False" -> $g
            fmt::autotype "FALSE" -> $h
            fmt::autotype "FALSE" -> $h
            fmt::autotype "-42" -> $i
            fmt::autotype "-3.1415" -> $j
            fmt::autotype "0" -> $k
            fmt::autotype "1" -> $l
        """))

    def test_strip(self):
        self.assertEqual({
            'a': 'Value',
            'b': 'Value',
            'c': 'Value',
            'd': 'Value',
            'e': 'Value',
            'f': 'Value',
            'g': '    Value',
            'h': 'Value    ',
            'i': 'Value!',
            'j': 'Value    ',
            'k': '    Value',
        }, self._eval("""
            fmt::strip 'Value' -> $a
            fmt::strip '    Value' -> $b
            fmt::strip 'Value    ' -> $c
            fmt::strip '  \n\nValue   ' -> $d
            fmt::strip begin

                Value
            end -> $e

            fmt::strip 'Value' {
                characters: '$',
            } -> $f

            fmt::strip '$    Value' {
                characters: '$',
            } -> $g

            fmt::strip 'Value    $$$' {
                characters: '$',
            } -> $h

            fmt::strip '$$#@$Value!$&#' {
                characters: '@#&$',
            } -> $i

            fmt::lstrip '    Value    ' -> $j
            fmt::rstrip '    Value    ' -> $k
        """))



from __future__ import absolute_import
import inspect
import logging
import re
import webfriend.scripting.commands
from webfriend.scripting.environment import Environment
from webfriend.scripting.scope import Scope
from webfriend.scripting.commands.base import CommandProxy
from webfriend.utils import get_module_from_string, resolve_object

RX_DOCSTRING_SEEOTHER = re.compile('(?P<head>.*)\s*see:\s*`(?P<module>.*)`\s*(?P<tail>.*)', re.IGNORECASE)


def document_commands(
    plugins=None,
    only_plugins=False,
    omit_header=False,
    oneshot=False,
    environment=None,
    commands=None
):
    base = CommandProxy

    if not environment:
        environment = Environment(Scope())

    toc = []
    subtoc = {}
    commands_body = []

    if isinstance(plugins, list):
        for plugin in plugins:
            environment.register_by_module_name(plugin)

    classes = environment.proxies

    classes = [
        i for q, i in classes.items() if q == CommandProxy.default_qualifier
    ] + sorted([
        i for q, i in classes.items() if q != CommandProxy.default_qualifier
    ])

    # for each proxy class
    for proxy in classes:
        if only_plugins:
            if plugins is None:
                continue

            if isinstance(plugins, list) and proxy.as_qualifier() not in plugins:
                continue

        proxy_docs = inspect.getdoc(proxy)

        if not oneshot:
            commands_body.append("## `{}` Command Set".format(proxy.as_qualifier()))
            commands_body.append('')

        # add top-level TOC item for this command set
        if proxy.doc_name:
            toc_title = proxy.doc_name
        else:
            toc_title = proxy.as_qualifier().title()

        tocentry = (toc_title, proxy.as_qualifier())

        if tocentry not in toc:
            toc.append(tocentry)

        subtoc[proxy.as_qualifier()] = set()

        # add command set documentation (if any)
        if proxy_docs and not oneshot:
            commands_body.append(proxy_docs)
            commands_body.append('')

        # for each direct method of this proxy (omitting private and inherited methods)
        for method in inspect.getmembers(proxy, inspect.ismethod):
            if isinstance(commands, list) and len(commands):
                if proxy.qualify(method[0]) not in commands:
                    continue

            if method[0].startswith('_'):
                continue

            if method[0] in dir(base):
                continue

            # get method documentation
            method_doc = get_docstring(method[1])

            commands_body.append("### `{}`".format(proxy.qualify(method[0])))

            # add method
            subtoc[proxy.as_qualifier()].add(
                (proxy.qualify(method[0]), (method_doc is not None))
            )

            commands_body.append("\n```\n{}{}\n```".format(
                proxy.qualify(''),
                get_signature(method)
            ))

            if method_doc:
                commands_body.append('')
                commands_body.append(method_doc)

            commands_body.append('')
            commands_body.append('---')
            commands_body.append('')

        commands_body.append('')

    if not oneshot:
        final = ['# Command Reference']
    else:
        final = []

    if not omit_header and not oneshot:
        header_docs = inspect.getdoc(webfriend.scripting.commands)

        if header_docs:
            final.append('')
            final += header_docs.split('\n')
            final.append('')

    # include TOC tree
    if not oneshot:
        for title, qualifier in toc:
            final.append("- [{}](#{}-command-set)".format(title, qualifier))

            if qualifier in subtoc:
                for method, has_docs in sorted(list(subtoc[qualifier])):
                    wrap_l = wrap_r = '**'

                    if not has_docs:
                        wrap_l = wrap_r = ''

                    final.append("   - {}[{}](#{}){}".format(wrap_l, method, method.replace('::', ''), wrap_r))

    # include command documentation lines
    final.append('')
    final += commands_body

    return final


def get_docstring(obj):
    docstring = inspect.getdoc(obj)

    if not docstring:
        return None

    out = []
    match = RX_DOCSTRING_SEEOTHER.match(docstring.strip())

    if match:
        matches = match.groupdict()
        out = matches.get('head', '')

        if len(matches.get('module', '')):
            module, remainder = get_module_from_string(matches['module'])

            if module:
                subobj = resolve_object(remainder, parent=module)

                if subobj:
                    subdocs = inspect.getdoc(subobj)

                    if subdocs:
                        out += subdocs
                    else:
                        logging.warning('Referenced object "{}" has no documentation.'.format(
                            matches['module']
                        ))
                else:
                    logging.warning('Failed to resolve "{}" into an object'.format(
                        matches['module']
                    ))
            else:
                logging.warning('Failed to resolve "{}" into a module'.format(
                    matches['module']
                ))

        out += matches.get('tail', '')

        docstring = out

    if not len(docstring):
        return None

    return docstring


def get_signature(method):
    arguments = []
    argspec = inspect.getargspec(method[1])
    args = list(argspec.args)
    args.reverse()
    maxarglen = max([len(a) for a in args]) + 1

    if argspec.defaults:
        defaults = list(argspec.defaults)
        defaults.reverse()
    else:
        defaults = []

    for i, arg in enumerate(args):
        skip_default = False

        if arg == 'self':
            continue

        if len(defaults) and i < len(defaults):
            default = defaults[i]
        else:
            default = '<REQUIRED>'

        if len(args) == 1:
            skip_default = True
        elif i == (len(args) - 2):
            skip_default = True
        else:
            arg = (arg + ':').ljust(maxarglen)

        if not skip_default:
            wrap_l = ''
            wrap_r = ''

            if default is True:
                default = 'true'

            elif default is False:
                default = 'false'

            elif default is None:
                default = 'null'

            elif default == '<REQUIRED>':
                default

            elif isinstance(default, basestring):
                wrap_l = "'"
                wrap_r = "'"

            arg = '    {} {}{}{}'.format(arg, wrap_l, default, wrap_r)

        arguments = [arg] + arguments

    if len(arguments) == 0:
        return method[0]
    elif len(arguments) == 1:
        return '{} <{}>'.format(
            method[0],
            arguments[0].upper()
        )
    else:
        return '{} <{}> {{\n{}\n}}'.format(
            method[0],
            arguments[0].upper(),
            ',\n'.join(arguments[1:])
        )


if __name__ == '__main__':
    logging.basicConfig(level=logging.WARNING)
    print('\n'.join(document_commands()))



from __future__ import absolute_import


class Event(object):
    nested_joiner = '.'

    def __init__(self, rpc, event, payload=None):
        self.rpc     = rpc
        self.event   = '{}.{}'.format(rpc.domain, event)
        self.payload = {}

        if isinstance(payload, dict):
            self.payload = payload

    def to_json(self):
        return {
            'event':   str(self),
            'payload': self.payload,
        }

    def get(self, key, fallback=None):
        parts = key.split(self.nested_joiner)

        for i, part in enumerate(parts):
            if isinstance(part, str):
                parts[i] = part.decode('UTF-8')

        base = self.payload

        for k in parts:
            if isinstance(base, (list, tuple)):
                try:
                    base = base[int(k)]
                    continue
                except ValueError:
                    return fallback

            # try to access key via the [] operator
            if isinstance(base, dict):
                try:
                    base = base[k]
                    continue
                except KeyError:
                    return fallback

        if base is None:
            return fallback

        return base

    def matches_criteria(self, criteria):
        for key, criterion in criteria.items():
            # regex matches
            if hasattr(criterion, 'match'):
                if criterion.match('{}'.format(self.get(key))):
                    continue
                else:
                    return False

            # exact matches
            if self.get(key) == criterion:
                continue
            else:
                return False

        return True

    def __getitem__(self, key):
        if key == '__payload__':
            return self.payload
        else:
            return self.get(key)

    def __contains__(self, key):
        if key == '__payload__':
            return True

        return (key in self.payload)

    def __repr__(self):
        return self.event

    def __str__(self):
        return self.__repr__()



from __future__ import absolute_import
import os
from sys import platform


def locate_browser_process():
    names = []

    if os.getenv('WEBFRIEND_BROWSER'):
        names += os.getenv('WEBFRIEND_BROWSER').split(':')

    if platform.startswith('linux'):
        names += [
            'chromium-browser',
            'google-chrome',
        ]

    elif platform == 'darwin':
        names += [
            '/Applications/Google Chrome Canary.app/Contents/MacOS/Google Chrome Canary',
            '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome',
            '/Applications/Chromium.app/Contents/MacOS/Chromium',
        ]

    # We'll get there...
    # elif platform == 'win32':
    #     names = []

    if not len(names):
        raise Exception("Unsupported platform '{}'".format(platform))

    for name in names:
        name = which(name)

        if name:
            return name

    raise Exception("Unable to locate browser process")


def which(base):
    if os.path.exists(base):
        return base

    for path in os.environ['PATH'].split(os.pathsep):
        if os.path.exists(os.path.join(path, base)):
            return os.path.join(path, base)

    return None



""" Module that monkey-patches json module when it's imported so
JSONEncoder.default() automatically checks for a special "to_json()"
method and uses it to encode the object if found.
"""
from json import JSONEncoder


def _default(self, obj):
    if hasattr(obj.__class__, 'to_json'):
        return obj.to_json()
    else:
        try:
            _default.default(obj)
        except TypeError:
            return obj.__repr__()
        except UnicodeEncodeError:
            return obj.__repr__()


_default.default = JSONEncoder().default  # Save unmodified default.
JSONEncoder.default = _default            # replacement



from __future__ import absolute_import
from __future__ import unicode_literals
import logging
import time
import io
from webfriend.scripting.commands.base import CommandProxy
from webfriend import exceptions


class PageProxy(CommandProxy):
    def screenshot(
        self,
        destination=None,
        width=0,
        height=0,
        x=-1,
        y=-1,
        format='png',
        jpeg_quality=None,
        selector=[],
        settle=250,
        after_events=None,
        settle_timeout=None,
        reply_timeout=30000,
        autoclose=True
    ):
        """
        Capture the current screen contents as an image and write the image to a file or return it
        as a file-like object.

        #### Arguments

        - **destination** (`str`, _file-like object_, optional):

            If given as a string, this will be the filesystem path that the image is written to.
            If given as a file-like object, that object will be written to, seeked back to zero,
            and returned.  If `None`, an `io.BytesIO` buffer will be allocated, written to, and
            returned.

        - **width**, **height** (`int`, optional):

            If given, this is the width and/or height the viewport will be resized to before
            capturing the image. If not specified, the dimensions of the element specified by
            'selector' will be queried and that element's outerWidth/outerHeight will be used.

        - **x**, **y** (`int`, optional):

            If given, this is the x/y position the viewport will be moved to.  If a negative number
            is specified, then the x/y coordinates of the element matched by **selector** will be
            used.

        - **format** (`str`):

            The output format of the image, either `png` or `jpeg`.

        - **jpeg_quality** (`int`, optional):

            If given, and if **format** is `jpeg`, this is the quality of the JPEG image (0-100).

        - **selector** (`str`, `list`):

            When detecting the width and/or height of the page area to render, this is the HTML
            element that will be measured.  Changing this to a selector that matches a specific
            element (using the default settings for **width**, **height**, **x**, and **y**) will
            result in taking a screenshot of _just_ that element.

            If a `list` is specified, all elements in the list will be retrieved and the one with
            the largest height will be used.

        - **settle** (`int`, optional):

            The number of milliseconds to wait after performing the viewport resize before
            actually capturing the image.  This gives newly-exposed elements time to load.

        - **after_events** (`str`, `list`, optional):

            If an event name or list of event names is given, then **settle** will be interpreted
            as an amount of time _after_ the last of these named events has been received. If the
            special value `any` is provided, then **settle** will wait for that amount of time
            since any event has been received to elapse, up to 'settle_timeout' seconds.

        - **settle_timeout** (`int`, optional):

            The maximum amount of time to wait for events to stop coming in as described in by
            **after_events**.

        - **reply_timeout** (`int`, optional):

            The maximum amount of time to wait for the screenshot call to complete.

        - **autoclose** (`bool`):

            If a file handle is given as the **destination**, should it be automatically closed
            when the screenshot is completed.

        #### Returns
        `dict`, with keys:

        - _element_ (`webfriend.rpc.DOMElement`):
            The element that was used as a measurement reference.

        - _width_ (`int`):
            The final width of the viewport that was captured.

        - _height_ (`int`):
            The final height of the viewport that was captured.

        - _x_ (`int`):
            The final X-position of the viewport that was captured.

        - _y_ (`int`):
            The final Y-position of the viewport that was captured.

        - _destination_ (`object`, optional):
            The destination file-like object data was written to, if specified.

        - _path_ (`str`, optional):
            The filesystem path of the file data was written to, if specified.
        """

        element = None
        max_element = None

        # width defaults to document scrollWidth
        if not width:
            width = self.tab.dom.scroll_width

        # height defaults to document scrollHeight
        if not height:
            height = self.tab.dom.scroll_height

        # listify selector
        if selector is None:
            selector = []
        elif not isinstance(selector, list):
            selector = [selector]

        # if selector is a list, then find the tallest element among all of them
        for s in selector:
            try:
                elements = self.tab.dom.query_all(s)
                element = self.tab.dom.ensure_unique_element(s, elements)

                if max_element:
                    if element.height > max_element.height:
                        max_element = element

                else:
                    max_element = element

            except (exceptions.EmptyResult, exceptions.TooManyResults):
                continue

        if max_element:
            element = max_element

        return_flo = True

        # ---
        # if an element was found and an explicit override was not set...
        if element and not width:
            width = element.scroll_width

        if element and not height:
            height = element.scroll_height

        if element and x < 0:
            x = element.left
        else:
            x = 0

        if element and y < 0:
            y = element.top
        else:
            y = 0
        # ---

        if width and height:
            # resize and force redraw
            self.tab.emulation.set_device_metrics_override(
                width=width,
                height=height,
                device_scale_factor=1.0
            )

            self.tab.emulation.set_visible_size(width, height)

        try:
            self.tab.emulation.force_viewport(x=x, y=y)
        except exceptions.ProtocolError:
            pass

        # wait for a spell for the page to adjust to its new world
        if settle:
            if after_events is None:
                logging.info('Waiting {}ms for resize to settle'.format(settle))
                time.sleep(settle / 1e3)
            else:
                wfi_params = {}

                if settle_timeout:
                    wfi_params['timeout'] = settle_timeout

                if after_events != 'any':
                    wfi_params['event_filter'] = after_events

                try:
                    logging.info('Waiting {}ms after last events'.format(settle))
                    self.tab.wait_for_idle(settle, **wfi_params)
                except exceptions.TimeoutError:
                    pass

        # setup a FLO to write data to if we don't have one
        if destination is None:
            destination = io.BytesIO()
        elif isinstance(destination, basestring):
            return_flo = False

        self.tab.page.capture_screenshot(destination, format=format, reply_timeout=reply_timeout)

        out = {
            'element': element,
            'width':   width,
            'height':  height,
            'x':       x,
            'y':       y,
        }

        if return_flo:
            destination.seek(0)
            out['destination'] = destination

            if autoclose:
                if hasattr(destination, 'close'):
                    close = getattr(destination, 'close')

                    if callable(close):
                        close()
        else:
            out['path'] = destination

        return out

    def start_capture(
        self,
        destination,
        duration=None,
        format='png',
        jpeg_quality=85,
        every_nth=None,
        filename_format=None
    ):
        self.tab.page.start_screencast(
            destination=destination,
            duration=duration,
            format=format,
            jpeg_quality=jpeg_quality,
            every_nth=every_nth,
            filename_format=filename_format,
            additional_values=self.scope.as_dict()
        )

    def wait_for_capture(self):
        self.tab.wait_for('Page.screencastVisibilityChanged')

    def stop_capture(self):
        self.tab.page.stop_screencast()

    def find(self, text, **kwargs):
        logging.info('text={}, kwargs={}'.format(text, kwargs))
        raise exceptions.NotImplemented("find()")

    def source(self, selector=None):
        """
        Retrieves the `outerHtml` of the matching page element.

        #### Arguments

        - **selector** (`str`):

            The page element to retrieve source for, given as a CSS-style selector, an ID
            (e.g. "#myid"), or an XPath query (e.g.: "xpath://body/p").

        #### Returns
        A `unicode` string representing the HTML contents of the matched element.

        #### Raises
        - `webfriend.exceptions.EmptyResult` if zero elements were matched, or
        - `webfriend.exceptions.TooManyResults` if more than one elements were matched.
        """
        if selector:
            elements = self.tab.dom.select_nodes(selector)
            self.tab.dom.ensure_unique_element(selector, elements)
            return elements['nodes'][0].outer_html
        else:
            return self.tab.dom.root.outer_html

    # def highlight(self, selector, **kwargs):
    #     elements = self.tab.dom.select_nodes(selector)
    #     self.tab.dom.ensure_unique_element(selector, elements)
    #     self.tab.overlay.highlight_node(node_id=elements['nodes'][0].id, **kwargs)

    def remove(self, selector):
        """
        Removes ALL elements that match the given selector from the page.

        #### Arguments

        - **selector** (`str`):

            The page elements to remove, given as a CSS-style selector, an ID
            (e.g. "#myid"), or an XPath query (e.g.: "xpath://body/p").

        #### Returns
        A `list` of `webfriend.rpc.DOMElement` instances that were matched and removed from the
        page.
        """
        removed = []

        for element in self.tab.dom.query_all(selector):
            self.tab.dom.remove(element.id)
            removed.append(element)

        return removed

    def dump_dom(self):
        """
        Print the portion of the DOM the local page cache is currently aware of to log output.
        This is primarily useful for debugging WebFriend.
        """
        self.tab.dom.print_node()

    def resources(self):
        """
        Return a list of objects describing every network request that has occurred since the last
        invocation of `[go](#go)`, since the resource cache was last cleared with
        `[page::clear_resources](#pageclear_resources)`, or since the current tab was loaded.

        #### Returns
        A `list` of `dict` containing details about the network requests that were performed.
        """
        return self.tab.dom.resources.values()

    def resource(self, url, **kwargs):
        """
        See: `webfriend.rpc.DOM.get_resource`
        """
        kwargs.update({
            'url': url,
        })

        request = self.tab.dom.get_resource(**kwargs)

        if request and request.get('completed'):
            out = {
                'request': request,
                'data': None,
            }

            try:
                out['data'] = io.BytesIO()
                out['data'].write(self.tab.network.get_response_body(request['id']))
                out['data'].seek(0)
            except:
                pass

            return out
        else:
            return False

    def dialog_cancel(self):
        """
        Cancels an open modal dialog (presses 'Cancel').
        """
        self.tab.page.dialog(False)

    def dialog_ok(self):
        """
        Accepts an open modal dialog (presses 'OK' or 'Yes').
        """
        self.tab.page.dialog(True)

    def prompt_text(self, text, submit=True):
        """
        Enters the given text into an open prompt dialog and (optionally) submits it.

        #### Arguments

        - **text** (`str`):

            The text value to enter into the prompt dialog.

        - **submit** (`bool`, optional):

            Whether to automatically submit the dialog or not.
        """
        self.tab.page.dialog(submit, text=text)



import os.path
import random
import inspect
import importlib
import string

PACKAGE_ROOT = os.path.abspath(
    os.path.dirname(
        os.path.dirname(__file__)
    )
)

PACKAGE_NAME = os.path.basename(PACKAGE_ROOT)


def random_string(count, charset=string.lowercase + string.digits):
    return ''.join(random.sample(charset, count))


def autotype(value):
    if isinstance(value, basestring):
        if value.lower() == 'true':
            return True
        elif value.lower() == 'false':
            return False
        else:
            try:
                return int(value)
            except ValueError:
                pass

            try:
                return float(value)
            except ValueError:
                pass

    return value


def get_module_from_string(string, package=None):
    parts = string.split('.')
    remainder = []

    while len(parts):
        try:
            return importlib.import_module('.'.join(parts), package=package), remainder
        except ImportError:
            remainder = [parts.pop()] + remainder

    return None, string.split('.')


def resolve_object(parts, parent=None):
    if not parent:
        parent = globals()

    while len(parts):
        proceed = False

        for member in inspect.getmembers(parent):
            if member[0] == parts[0]:
                parent = member[1]
                parts = parts[1:]
                proceed = True
                break

        if not proceed:
            return None

    return parent



"""
This module contains documentation on all of the commands supported by the WebFriend core
distribution.  Additional commands may be included via plugins.

## Documentation

All commands are documented by describing the format for executing the command.  All supported
options are presented, along with their default values.  If an option is required, it will be
shown with the value `<REQUIRED>`.
"""

from __future__ import absolute_import
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)  # noqa



from __future__ import absolute_import
import click
import click_log
import json
import logging
import os
import sys
import signal
import traceback
import webfriend.info
from webfriend.browser import Chrome
from webfriend.scripting.parser.exceptions import UserError
from webfriend.scripting.environment import Environment
from webfriend.scripting.execute import execute_script
from webfriend.repl import REPL
from webfriend.utils.docs import document_commands

log = logging.getLogger()
err = logging.StreamHandler(stream=sys.stderr)
err.setFormatter(click_log.ColorFormatter())
log.addHandler(err)


@click.group(invoke_without_command=True)
@click.option('--interactive', '-I', is_flag=True, default=False)
@click.option('--suppress-output', '-o', is_flag=True, default=False)
@click.option('--no-temp-profile', '-T', is_flag=True, default=False)
@click.option('--generate-docs', is_flag=True, default=False)
@click.option('--only-document-plugins', is_flag=True, default=False)
@click.option('--omit-header', is_flag=True, default=False)
@click.option('--debug', '-D', is_flag=True, default=False)
@click.option('--version', '-v', is_flag=True, default=False)
@click.option(
    '--debugger-url', '-u',
    metavar='URL',
    help='The URL of the Chrome Remote Debugger to connect to.'
)
@click.option(
    '--log-level',
    '-L',
    default='WARNING'
)
@click.option(
    '--script-log-level',
    '-S',
    default='DEBUG'
)
@click.option(
    '--plugins', '-p',
    metavar='PLUGIN[,PLUGIN ..]',
    help='A comma-separated list of additional plugins to load'
)
@click.argument('script', type=click.File('rb'), required=False)
@click.argument('remainder', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def main(
    ctx,
    interactive,
    suppress_output,
    no_temp_profile,
    generate_docs,
    only_document_plugins,
    omit_header,
    debug,
    version,
    debugger_url,
    log_level,
    script_log_level,
    plugins,
    script,
    remainder
):
    if version is True:
        print('webfriend {}'.format(webfriend.info.version))
        sys.exit(0)

    log.setLevel(logging.getLevelName(log_level.upper()))

    if plugins:
        plugins = plugins.split(',')

    if generate_docs:
        click.echo('\n'.join(document_commands(
            plugins=plugins,
            only_plugins=only_document_plugins,
            omit_header=omit_header
        )))
    else:
        if debug:
            os.environ['WEBFRIEND_DEBUG'] = 'true'

        if not interactive and script is None and sys.stdin.isatty():
            raise IOError("Must provide the path to script to execute.")

        # using the with-syntax launches an instance of chrome in the background before proceeding
        with Chrome(
            debug_url=debugger_url,
            use_temp_profile=(not no_temp_profile)
        ) as chrome:
            environment = Environment(browser=chrome, log_level=script_log_level)

            if isinstance(plugins, list):
                for plugin in plugins:
                    environment.register_by_module_name(plugin)

            # interactive REPL
            if interactive:
                repl = REPL(chrome, environment=environment)
                repl.run()
            else:
                # by now we've validated that there is not a TTY waiting at stdin, so there
                # must be some data to read.  In that case, read the script from stdin.
                #
                # This let's users call `webfriend` by piping commands to it or by passing in heredocs
                #
                if script is None:
                    script = sys.stdin

                # execute script as file (or via shebang invocation)
                scope = execute_script(
                    chrome,
                    script.read(),
                    environment=environment
                ).as_dict()

            if not suppress_output:
                click.echo(json.dumps(scope, indent=4))


try:
    def signal_handler(signal, frame):
        log.error('Interrupted')
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    main()
except UserError as e:
    exit(e.exit_code)
except Exception as e:
    log.debug(traceback.format_exc())
    log.error(str(e))
    exit(1)



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.scope import Scope
from webfriend.scripting import parser
from webfriend.scripting.commands.base import CommandProxy
import sys
import logging
import traceback
import inspect
import importlib
import colorlog


class EnvironmentNotReady(Exception):
    pass


class IgnoreResults(object):
    pass


class Environment(object):
    default_result_key = 'result'

    def __init__(self, scope=None, proxies=None, browser=None, log_level='DEBUG'):
        self._scope = (scope or Scope())
        self.proxies = (proxies or {})
        self.script = None
        self.registered_names = set()
        self.browser = browser
        self._is_ready = False
        self._exec_options = {}
        self.register_defaults()
        self.sync_scopes()
        self.ready()
        self._line = 0
        self._col  = 0
        self._log_handler = colorlog.StreamHandler(stream=sys.stderr)
        self._log_handler.setFormatter(colorlog.ColoredFormatter('%(log_color)s%(message)s'))
        self.log = colorlog.getLogger('friendscript')
        self.log.propagate = False
        self.log.setLevel(logging.getLevelName(log_level.upper()))
        self.log.addHandler(self._log_handler)

    @property
    def scope(self):
        return self._scope

    @property
    def line(self):
        return self._line

    @property
    def col(self):
        return self._col

    def isolated(self, scope=None):
        return Environment(
            scope=scope,
            proxies=self.proxies,
            browser=self.browser
        )

    def ready(self):
        if self._is_ready:
            return True

        if self.browser:
            self.browser.default.enable_events()

        self._is_ready = True
        return True

    def sync_scopes(self):
        for _, proxy in self.proxies.items():
            proxy.scope = self._scope

    def set_scope(self, scope):
        self._scope = scope
        self.sync_scopes()

    def get_command_names(self):
        names = []

        for _, proxy in self.proxies.items():
            names += proxy.get_command_names()

        # sort the list of by longest-to-shortest
        names.sort(key=len, reverse=True)

        return names

    def register_defaults(self):
        for name, cls in CommandProxy.get_all_proxies():
            self.register(cls, qualifier=name)

    def register_by_module_name(
        self,
        proxy_module_name,
        package_format='webfriend.scripting.commands.{}'
    ):
        plugin_module_name = package_format.format(proxy_module_name)

        # only register this module if we haven't already, making this call idempotent for
        # modules named the same thing
        if plugin_module_name not in self.registered_names:
            module = importlib.import_module(plugin_module_name)

            for name, obj in inspect.getmembers(module, inspect.isclass):
                if issubclass(obj, CommandProxy) and obj is not CommandProxy:
                    self.register(obj)
                    self.registered_names.add(plugin_module_name)

    def register(self, proxy_cls, qualifier=None):
        if not issubclass(proxy_cls, CommandProxy):
            raise ValueError("Proxy class must be a subclass of CommandProxy")

        if qualifier is None:
            qualifier = proxy_cls.as_qualifier()

        logging.debug('Registering {} as command proxy with qualifier "{}"'.format(proxy_cls, qualifier))
        self[qualifier] = proxy_cls(self.browser, environment=self)

    def get_proxy_for_command(self, command):
        q_command = command.name.split('::')
        proxy = self

        try:
            if len(q_command) == 1:
                if q_command[0] in proxy and proxy[q_command[0]].has_default():
                    proxy = proxy[q_command[0]]
                else:
                    proxy = proxy[CommandProxy.default_qualifier]

                command_name = q_command[0]
            else:
                for subproxy in q_command[:-1]:
                    proxy = proxy[subproxy]

                command_name = q_command[-1]

            return proxy, command_name
        except KeyError:
            raise parser.exceptions.ScriptError("Unknown command '{}'".format(command.name))

    def has_execution_option(self, key):
        return key in self._exec_options

    def get_execution_option(self, key, fallback=None):
        return self._exec_options.get(key, fallback)

    def set_execution_option(self, key, value):
        self._exec_options[key] = value

    def clear_execution_option(self, key):
        try:
            del self._exec_options[key]
        except KeyError:
            pass

    def prepare_output_values(self, values):
        if isinstance(values, list):
            return [self.prepare_output_values(v) for v in values]

        elif isinstance(values, dict):
            return dict([
                (k, self.prepare_output_values(v)) for k, v in values.items()
            ])

        elif values is True:
            return 'true'

        elif values is False:
            return 'false'

        elif values is None:
            return ''

        return values

    def interpolate(self, value, scope=None, deindent_heredocs=True, **kwargs):
        if scope is None:
            scope = self._scope

        scope = Scope(kwargs, parent=scope)
        scopevars = scope.as_dict()

        if isinstance(value, parser.lang.Expression):
            value, _ = value.process(scope, preserve_types=True)

        # if this is an exact-match string, then interpolate is a no-op
        if isinstance(value, parser.types.StringLiteral):
            if isinstance(value, str):
                value = value.decode('UTF-8')

            return value

        # for heredocs, we're going to return the literal as-is, but we trim off the indentation
        # whitespace of the calling command, that way heredocs can be written to match the indentation
        # level of the surrounding script, but when evaluated they are indented to the "intended" level
        if isinstance(value, parser.types.Heredoc):
            if deindent_heredocs:
                return value.deindent(self.col - 1)
            else:
                return value

        actual = parser.to_value(value, scope)

        if isinstance(actual, str):
            actual = actual.decode('UTF-8')

        # lists get recursed into
        if isinstance(actual, list):
            # recurse into lists
            return [self.interpolate(v, **scopevars) for v in actual]

        # dicts also get recursed into
        elif isinstance(actual, dict):
            # recurse into dicts
            return dict([
                (k, self.interpolate(v, **scopevars)) for k, v in actual.items()
            ])

        # strings (as returned from to_value) get interpolated
        elif isinstance(actual, basestring):
            # do the interpolation
            return actual.format(**self.prepare_output_values(scopevars))

        # everything else passes through
        else:
            return actual

    def execute(self, command, scope=None):
        if scope:
            # NOTE: Future Sadness May Await
            #
            # this here is why commands should not execute in parallel, because we're changing the scope
            # of the _entire_ environment to the given scope for the duration of this call
            self.set_scope(scope)

        # if the script manager is specified, use it to record the current line/col we are evaluating
        # in the script
        self._line, self._col = command._tx_metamodel.parser.pos_to_linecol(
            command._tx_position
        )

        # get the proxy and proxy-local command name from the CommandSet
        proxy, command_name = self.get_proxy_for_command(command)

        # the proxy needs to know about the command
        if command_name in proxy:
            fn = proxy[command_name]

            # interpolate option values
            if command.options is not None:
                opts = command.options.as_dict()
            else:
                opts = {}

            try:
                # recursively interpolate all options
                opts = self.interpolate(opts)
                keyname = None
                command_id = command.id

                if command_id is not None:
                    if command_id.variable:
                        command_id = command_id.resolve(scope)
                    else:
                        # interpolate command ID
                        command_id = self.interpolate(command_id.value)

                # figure out where we want to store results
                if isinstance(command.resultkey, parser.variables.Variable):
                    resultkey = command.resultkey.as_key(scope)
                    keyname = '.'.join(resultkey)

                elif isinstance(command.resultkey, basestring) and len(command.resultkey):
                    resultkey = self.interpolate(command.resultkey)

                else:
                    resultkey = self.default_result_key

                if resultkey is not None:
                    keyname = '$' + (keyname or str(resultkey))
                else:
                    keyname = '[discarded]'

            except KeyError as e:
                raise parser.exceptions.ScriptError(
                    "Use of undefined variable '{}' in string pattern".format(
                        str(e).replace("u'", '', 1).rstrip("'")
                    ),
                    model=command
                )

            logging.debug(' ========= Execute: {} -> {}'.format(
                proxy.qualify(command_name),
                keyname
            ))

            # call function
            try:
                if command_id is None:
                    return resultkey, fn(**opts)
                else:
                    return resultkey, fn(
                        command_id,
                        **opts
                    )
            except Exception as e:
                line = col = None

                if hasattr(e, 'line'):
                    line = e.line

                if hasattr(e, 'col'):
                    col = e.col

                logging.debug(traceback.format_exc(e))

                raise parser.exceptions.CommandExecutionError(
                    "Error running command '{}': {}".format(command_name, e),
                    model=command,
                    line=line,
                    col=col
                )

        else:
            raise parser.exceptions.ScriptError("No such command '{}'".format(command_name), model=command)

    def __contains__(self, proxy_name):
        if proxy_name in self.proxies:
            return True

        return False

    def __getattr__(self, proxy_name):
        return self.proxies[proxy_name]

    def __getitem__(self, proxy_name):
        if not self._is_ready:
            raise EnvironmentNotReady("Cannot work with environment until ready() is called.")

        return self.proxies[proxy_name]

    def __setitem__(self, proxy_name, proxy):
        self.proxies[proxy_name] = proxy
        self.sync_scopes()



from __future__ import absolute_import
from webfriend.rpc import Base


class Console(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Console
    """
    domain = 'Console'



from __future__ import absolute_import
from webfriend.rpc import Base
import time
from random import random


class Input(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Input
    """

    domain = 'Input'
    supports_events = False
    key_event_types = ['keyDown', 'keyUp', 'rawKeyDown', 'char']
    key_events_with_text = ['keyDown', 'char']
    key_args = [
        ('timestamp', 'timestamp'),
        ('unmodified_text', 'unmodifiedText'),
        ('key_identifier', 'keyIdentifier'),
        ('code', 'code'),
        ('dom_key', 'key'),
        ('windows_virtual_keycode', 'windowsVirtualKeyCode'),
        ('native_virtual_keycode', 'nativeVirtualKeyCode'),
        ('is_autorepeat', 'autoRepeat'),
        ('is_keypad', 'isKeypad'),
        ('is_system', 'isSystemKey'),
    ]

    mouse_event_types = ['mousePressed', 'mouseReleased', 'mouseMoved']
    mouse_buttons = ['left', 'middle', 'right']

    def ignore_input(self):
        self.call('setIgnoreInputEvents', ignore=True)

    def accept_input(self):
        self.call('setIgnoreInputEvents', ignore=False)

    def dispatch_key_event(
        self,
        event_type='keyDown',
        alt=False,
        control=False,
        meta=False,
        shift=False,
        timestamp=None,
        text=None,
        unmodified_text=None,
        key_identifier=None,
        code=None,
        dom_key=None,
        windows_virtual_keycode=None,
        native_virtual_keycode=None,
        is_autorepeat=False,
        is_keypad=False,
        is_system=False
    ):
        """
        See: https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-dispatchKeyEvent
        """
        params = {
            'type':      event_type,
            'modifiers': 0,
        }

        if event_type not in self.key_event_types:
            raise ValueError("event_type must be one of: {}".format(
                ', '.join(self.key_event_types)
            ))

        if event_type in self.key_events_with_text:
            if text is None:
                raise ValueError("text must be specified for '{}' event".format(event_type))
            else:
                params['text'] = text

        params['modifiers'] |= 1 if alt else 0
        params['modifiers'] |= 2 if control else 0
        params['modifiers'] |= 4 if meta else 0
        params['modifiers'] |= 8 if shift else 0

        for local, param_name in self.key_args:
            if locals()[local]:
                params[param_name] = locals()[local]

        self.call('dispatchKeyEvent', **params)

    def type_text(
        self,
        text,
        file=False,
        alt=False,
        control=False,
        meta=False,
        shift=False,
        is_keypad=False,
        key_down_time=30,
        key_down_jitter=0,
        delay=50,
        delay_jitter=0
    ):
        """
        Input the given textual data as keyboard input into the browser in its current state.

        #### Arguments

        - **text** (`str`, optional):

            The text string to input, one character or symbol at a time.

        - **file** (`str`, optional):

            If specified, read the text to input from the named file.

        - **alt**, **control**, **meta**, **shift** (`bool`):

            Declares that the Alt, Control, Meta/Command, and/or Shift keys (respectively) are
            depressed at the time of the click action.

        - **is_keypad** (`bool`, optional):

            Whether the text being input is issued via the numeric keypad or not.

        - **key_down_time** (`int`, optional):

            How long, in milliseconds, that each individual keystroke will remain down for.

        - **key_down_jitter** (`int`, optional):

            An amount of time, in milliseconds, to randomly vary the **key_down_time** duration
            from within each keystroke.

        - **delay** (`int`, optional):

            How long, in milliseconds, to wait between issuing individual keystrokes.

        - **delay_jitter** (`int`, optional):

            An amount of time, in milliseconds, to randomly vary the **delay** duration
            from between keystrokes.

        #### Returns
        The text that was submitted.
        """
        if file:
            text = open(text, 'r').read()
        elif text is None:
            raise ValueError("Must specify either 'text' or 'file' to read input from.")

        for char in text:
            self.dispatch_key_event(
                event_type='keyDown',
                text=char,
                alt=alt,
                control=control,
                meta=meta,
                shift=shift,
                is_keypad=is_keypad
            )

            # simulate the duration of the keypress
            if key_down_time:
                time.sleep((key_down_time * key_down_jitter * random()) / 1e3)

            # send the keyUp event
            self.dispatch_key_event(event_type='keyUp')

            # simulate the time between key presses
            if delay:
                time.sleep((delay * random() * delay_jitter) / 1e3)

        return text

    def dispatch_mouse_event(
        self,
        x,
        y,
        event_type='mouseMoved',
        button=None,
        alt=False,
        control=False,
        meta=False,
        shift=False,
        timestamp=None,
        click_count=1,
    ):
        """
        See https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-dispatchMouseEvent
        """
        params = {
            'type':      event_type,
            'x':         int(x),
            'y':         int(y),
            'modifiers': 0,
        }

        if event_type not in self.mouse_event_types:
            raise ValueError("event_type must be one of: {}".format(
                ', '.join(self.mouse_event_types)
            ))

        if button and button not in self.mouse_buttons:
            raise ValueError("button must be one of: {}".format(
                ', '.join(self.mouse_buttons)
            ))

        params['modifiers'] |= 1 if alt else 0
        params['modifiers'] |= 2 if control else 0
        params['modifiers'] |= 4 if meta else 0
        params['modifiers'] |= 8 if shift else 0

        if timestamp:
            params['timestamp'] = timestamp

        if button:
            params['button'] = button

        if click_count:
            params['clickCount'] = click_count

        self.call('dispatchMouseEvent', **params)

    def click_at(
        self,
        x,
        y,
        button='left',
        alt=False,
        control=False,
        meta=False,
        shift=False,
        move_to=True,
        drag=False,
        start_x=None,
        start_y=None,
        drag_start_delay=50,
        move_finish_delay=50,
        click_duration=50
    ):
        """
        Dispatch low-level mouse events at the given coordinates to perform a variety of actions.

        #### Arguments

        - **x**, **y** (`int`):

            The X- and Y-coordinates describing the click location.

        **button** (`str`):

            Which mouse button the event will be for; one of: `left`, `middle`, or `right`.

        **alt**, **control**, **meta**, **shift** (`bool`):

            Declares that the Alt, Control, Meta/Command, and/or Shift keys (respectively) are
            depressed at the time of the click action.

        **move_to** (`bool`):

            Simulates the mouse moving from the starting location to the destination **x**/**y**
            coordinates.

        **drag** (`bool`):

            If specified, the click operation will be a "drag and drop" style movement, wherein
            a set of initial coordinates receive a mouse press, then (with the button still
            pressed) the mouse will move to the destination coordinates, and _then_ the button
            release event will fire.

        **start_x** (`int`):

            For **drag** operations, the starting X-coordinate.

        **start_y** (`int`):

            For **drag** operations, the starting Y-coordinate.

        **drag_start_delay** (`int`):

            For **drag** operations, how long to wait (in milliseconds) after the initial mouse
            button press before starting to move.

        **move_finish_delay** (`int`):

            If simulating mouse movement with **move_to**, how long to wait (in milliseconds)
            _after_ the move is completed, but before clicking or releasing.

        **click_duration** (`int`):

            If non-zero and not dragging, how long (in milliseconds) the mouse button should be
            held for when  performing the click; i.e.: how long between the mouse press and mouse
            release events.
        """

        is_dragging = False

        # dragging entails pressing the mouse, moving it, then releasing it
        if drag:
            if start_x is not None and start_y is not None:
                self.dispatch_mouse_event(
                    start_x,
                    start_y,
                    event_type='mousePressed',
                    button=button,
                    alt=alt,
                    control=control,
                    meta=meta,
                    shift=shift,
                )

                is_dragging = True
                time.sleep(drag_start_delay / 1e3)

            else:
                raise AttributeError("If drag is specified, start_x and start_y must be given as well")

        # move the mouse, optionally with a button pressed if we're dragging
        if move_to:
            self.dispatch_mouse_event(
                x,
                y,
                event_type='mouseMoved',
                button=(button if is_dragging else None),
                alt=alt,
                control=control,
                meta=meta,
                shift=shift,
            )

            time.sleep(move_finish_delay / 1e3)

        # if we're dragging, then release the button now that we're where we're going
        if is_dragging:
            self.dispatch_mouse_event(
                x,
                y,
                event_type='mouseReleased',
                button=button,
                alt=alt,
                control=control,
                meta=meta,
                shift=shift,
            )

            is_dragging = False
        else:
            # otherwise, click the button
            self.dispatch_mouse_event(
                x,
                y,
                event_type='mousePressed',
                button=button,
                alt=alt,
                control=control,
                meta=meta,
                shift=shift,
            )

            if click_duration:
                # hold it.....
                time.sleep(click_duration / 1e3)

            # and release
            self.dispatch_mouse_event(
                x,
                y,
                event_type='mouseReleased',
                button=button,
                alt=alt,
                control=control,
                meta=meta,
                shift=shift,
            )



from __future__ import absolute_import
from webfriend.rpc import Base


class Schema(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Schema
    """
    domain = 'Schema'

    def get_domains(self):
        return self.call('getDomains')



from __future__ import absolute_import
from webfriend.rpc import Base


class Target(Base):
    """
    See: https://chromedevtools.github.io/devtools-protocol/tot/Target
    """
    domain = 'Target'
    supports_events = False

    def enable_auto_attach(self, wait_for_debugger=True):
        self.call('setAutoAttach', autoAttach=True, waitForDebuggerOnStart=wait_for_debugger)

    def disable_auto_attach(self):
        self.call('setAutoAttach', autoAttach=False)

    def enable_discover_targets(self):
        self.call('setDiscoverTargets', discover=True)

    def disable_discover_targets(self):
        self.call('setDiscoverTargets', discover=False)

    def enable_attach_to_frames(self):
        self.call('setAttachToFrames', value=True)

    def disable_attach_to_frames(self):
        self.call('setAttachToFrames', value=False)

    def send_message_to_target(self, target_id, message):
        self.call('sendMessageToTarget', targetId=target_id, message=message)

    def get_targets(self):
        return self.call('getTargets')

    def get_target_info(self, target_id):
        return self.call('getTargetInfo', targetId=target_id)

    def create_target(self, url, width=None, height=None, browser_context_id=None):
        params = {
            'url': url,
        }

        if width:
            params['width'] = width

        if height:
            params['height'] = height

        if browser_context_id:
            params['browserContextId'] = browser_context_id

        return self.call('createTarget', **params)

    def activate_target(self, target_id):
        self.call('activateTarget', targetId=target_id)

    def close_target(self, target_id):
        self.call('closeTarget', targetId=target_id)

    def attach_to_target(self, target_id):
        return self.call_boolean_response('attachToTarget', field='success', targetId=target_id)

    def detach_from_target(self, target_id):
        self.call('detachFromTarget', targetId=target_id)

    def create_browser_context(self):
        return self.call('createBrowserContext')

    def dispose_browser_context(self, browser_context_id):
        return self.call_boolean_response(
            'disposeBrowserContext',
            field='success',
            browserContextId=browser_context_id
        )



from __future__ import absolute_import
from __future__ import unicode_literals
from textx.model import model_root
from termcolor import colored
import re


class ScriptError(Exception):
    context_lines_before = 3
    context_lines_after = 3
    error_type = 'Error'

    def __init__(
        self,
        message,
        model=None,
        line=None,
        col=None,
        filename=None,
        data=None,
        **kwargs
    ):
        self.model = model
        self.line = line
        self.col = col
        self.filename = None

        # populate the lines list
        if isinstance(filename, basestring):
            self.filename = filename
            self.lines = open(filename).read().split('\n')

        elif isinstance(data, basestring):
            self.lines = data.split('\n')

        elif isinstance(model, object) and hasattr(model, '_tx_position'):
            root = model_root(model)

            if root and root._tx_metamodel:
                # determine line number and line position from absolute offsets
                line, col = root._tx_metamodel.parser.pos_to_linecol(
                    model._tx_position
                )

                if self.line is None:
                    self.line = line

                if self.col is None:
                    self.col = col

                # get the local Friendscript manager class from the root model,
                # then get the raw script data from there
                if root.manager and root.manager.data:
                    self.lines = root.manager.data.split('\n')
        else:
            self.lines = None

        self.message = self.prepare_message(message)

        super(ScriptError, self).__init__(self.message)

    def prepare_message(self, message):
        offending_char = None

        try:
            if self.lines and self.line and self.col:
                if self.line < len(self.lines):
                    offending_char = self.lines[self.line - 1][self.col - 1]
        except:
            pass

        # handles textx error messages that are not particularly friendly for end users
        if message.startswith('Expected '):
            if offending_char:
                message = "Unexpected character '{}'".format(offending_char)
            else:
                message = "Unexpected character"

        # script-aware traceback
        if self.lines and self.line and self.col:
            out = "{} on line {}, char {}\n".format(self.error_type, self.line, self.col)

            if self.filename:
                out += "  in file: {}\n".format(self.filename)

            if self.line <= len(self.lines):
                out += "Source:\n"
                start_line = (self.line - self.context_lines_before)
                end_line = (self.line + self.context_lines_after)

                if start_line < 1:
                    start_line = 1

                if len(self.lines) > 1:
                    if end_line >= len(self.lines):
                        end_line = len(self.lines) - 1
                else:
                    end_line = self.line

                for i in range(start_line, end_line + 1):
                    prefix = '   '
                    color = 'blue'
                    attrs = ['bold']

                    if i == self.line:
                        prefix = '>> '
                        color = 'yellow'
                        attrs = ['bold']

                    out += colored("{}{:4d}: {}\n".format(
                        prefix,
                        i,
                        self.lines[i - 1]
                    ), color, attrs=attrs)

                    if i == self.line:
                        out += colored("{}      {}{} {}\n".format(
                            prefix,
                            ' ' * (self.col - 1),
                            '^',
                            re.sub(r' at position.*', '', message)
                        ), 'red', attrs=attrs)
        else:
            out = "Syntax Error: {}".format(message)

        return out

    def __str__(self):
        return self.message


class SyntaxError(ScriptError):
    error_type = 'Syntax error'


class CommandExecutionError(ScriptError):
    error_type = 'Runtime error'


class UserError(Exception):
    error_type = 'User error'

    def __init__(self, message, exit_code=1, **kwargs):
        self.exit_code = exit_code
        super(UserError, self).__init__(message, **kwargs)



from __future__ import absolute_import
from __future__ import unicode_literals
from webfriend.scripting.parser import MetaModel


class CommandStanza(object):
    def __init__(self, parent, options):
        self.parent     = parent
        self.options    = (options or {})

    def as_dict(self):
        out = {}

        for kv in self.options:
            if not len(kv.values):
                continue

            if len(kv.values) == 1:
                out[kv.name] = kv.values[0]
            else:
                out[kv.name] = kv.values

        return out


class CommandSequence(MetaModel):
    pass
